#!/usr/bin/env python
"""
MESA Isochrones from:
http://waps.cfa.harvard.edu/MIST/iso_form.php
"""
import os
import sys
import glob
import copy
import tempfile
import subprocess
import shutil
from collections import OrderedDict as odict

try:
    from urllib.parse import urlencode
    from urllib.request import urlopen, Request
except ImportError:
    from urllib import urlencode
    from urllib2 import urlopen, Request

import numpy as np

from ugali.utils.logger import logger

from ugali.isochrone.parsec import Isochrone
from ugali.isochrone.model import get_iso_dir

###########################################################
# MESA Isochrones
# http://waps.cfa.harvard.edu/MIST/iso_form.php

# survey system
dict_output = odict([
        ('des','DECam'),
        ('sdss','SDSSugriz'),
        ('ps1','PanSTARRS'),
	    ('lsst','LSST'),
])

mesa_defaults = {
        'version':'1.0',
        'v_div_vcrit':'vvcrit0.4',
        'age_scale':'linear',
        'age_type':'single',
        'age_value':10e9, # yr if scale='linear'; log10(yr) if scale='log10'
        'age_range_low':'',
        'age_range_high':'',
        'age_range_delta':'',
        'age_list':'',
        'FeH_value':-3.0,
        'theory_output':'basic',
        'output_option':'photometry',
        'output':'DECam',
        'Av_value':0,
}

mesa_defaults_10 = dict(mesa_defaults,version='1.0')

class Dotter2016(Isochrone):
    """ MESA isochrones from Dotter 2016:
    http://waps.cfa.harvard.edu/MIST/interp_isos.html
    """
    _dirname =  os.path.join(get_iso_dir(),'{survey}','dotter2016')

    defaults = (Isochrone.defaults) + (
        ('dirname',_dirname,'Directory name for isochrone files'),
        ('hb_stage',3,'Horizontal branch stage name'),
        ('hb_spread',0.1,'Intrinisic spread added to horizontal branch'),
        )

    download_url = 'http://waps.cfa.harvard.edu/MIST'
    download_defaults = copy.deepcopy(mesa_defaults_10)

    abins = np.arange(1., 13.5+0.1, 0.1)
    zbins = np.arange(1e-5, 1e-3+1e-5, 1e-5)

    columns = dict(
            des = odict([
                (2, ('mass_init',float)),
                (3, ('mass_act',float)),
                (6, ('log_lum',float)),
                (9, ('u',float)),
                (10,('g',float)),
                (11,('r',float)),
                (12,('i',float)),
                (13,('z',float)),
                (14,('Y',float)),
                (15,('stage',float))
                ]),
            sdss = odict([
                (2, ('mass_init',float)),
                (3, ('mass_act',float)),
                (6, ('log_lum',float)),
                (9, ('u',float)),
                (10,('g',float)),
                (11,('r',float)),
                (12,('i',float)),
                (13,('z',float)),
                (14,('stage',float))
                ]),
            ps1 = odict([
                (2, ('mass_init',float)),
                (3, ('mass_act',float)),
                (6, ('log_lum',float)),
                (9, ('g',float)),
                (10,('r',float)),
                (11,('i',float)),
                (12,('z',float)),
                (13,('y',float)),
                (16,('stage',float))
                ]),
            lsst = odict([
                (2, ('mass_init',float)),
                (3, ('mass_act',float)),
                (6, ('log_lum',float)),
                (9, ('u',float)),
                (10,('g',float)),
                (11,('r',float)),
                (12,('i',float)),
                (13,('z',float)),
                (14,('Y',float)),
                (15,('stage',float))
                ]),
            )

    def _parse(self,filename):
        """
        Reads an isochrone in the Dotter 2016 format and determines
        the age (Gyr), metallicity (Z), and creates arrays with the
        initial stellar mass and corresponding magnitudes for each
        step along the isochrone.
        """
        try:
            columns = self.columns[self.survey.lower()]
        except KeyError as e:
            logger.warning('Unrecognized survey: %s'%(self.survey))
            raise(e)

        kwargs = dict(comments='#',usecols=list(columns.keys()),dtype=list(columns.values()))
        data = np.genfromtxt(filename,**kwargs)

        self.mass_init = data['mass_init']
        self.mass_act  = data['mass_act']
        self.luminosity = 10**data['log_lum']
        self.mag_1 = data[self.band_1]
        self.mag_2 = data[self.band_2]
        self.stage = data['stage']
        
        # Check where post-AGB isochrone data points begin
        self.mass_init_upper_bound = np.max(self.mass_init)
        self.index = np.nonzero(self.stage >= 4)[0][0]

        self.mag = self.mag_1 if self.band_1_detection else self.mag_2
        self.color = self.mag_1 - self.mag_2


    @classmethod
    def z2feh(cls, z):
        # Section 3.1 of Choi et al. 2016 (https://arxiv.org/abs/1604.08592)
        Z_init  = z                # Initial metal abundance
        Y_p     = 0.249            # Primordial He abundance (Planck 2015)
        c       = 1.5              # He enrichment ratio 

        Y_init = Y_p + c * Z_init 
        X_init = 1 - Y_init - Z_init

        Z_solar = 0.0142           # Solar metal abundance
        Y_solar = 0.2703           # Solar He abundance (Asplund 2009)
        X_solar = 1 - Y_solar - Z_solar

        return np.log10( Z_init/Z_solar * X_solar/X_init)

    @classmethod
    def feh2z(cls, feh):
        # Section 3.1 of Choi et al. 2016 (https://arxiv.org/abs/1604.08592)
        Y_p     = 0.249            # Primordial He abundance (Planck 2015)
        c       = 1.5              # He enrichment ratio 

        Z_solar = 0.0142           # Solar metal abundance
        Y_solar = 0.2703           # Solar He abundance (Asplund 2009)
        X_solar = 1 - Y_solar - Z_solar

        return (1 - Y_p)/( (1 + c) + (X_solar/Z_solar) * 10**(-feh))

    def query_server(self, outfile, age, metallicity):
        z = metallicity
        feh = self.z2feh(z)
        
        params = dict(self.download_defaults)
        params['output'] = dict_output[self.survey]
        params['FeH_value'] = feh
        params['age_value'] = age * 1e9
        if params['age_scale'] == 'log10':
            params['age_value'] = np.log10(params['age_value'])

        server = self.download_url
        url = server + '/iso_form.php'
        # First check that the server is alive
        logger.debug("Accessing %s..."%url)
        urlopen(url,timeout=2)

        #response = requests.post(url,data=params)
        q = urlencode(params).encode('utf-8')
        request = Request(url,data=q)
        response = urlopen(request)
        try:
            fname = os.path.basename(str(response.read()).split('"')[1])
        except Exception as e:
            logger.debug(str(e))
            msg = 'Output filename not found'
            raise RuntimeError(msg)
            
        tmpdir = os.path.dirname(tempfile.NamedTemporaryFile().name)
        tmpfile = os.path.join(tmpdir,fname)

        out = '{0}/tmp/{1}'.format(server, fname)
        cmd = 'wget --progress dot:binary %s -P %s'%(out,tmpdir)
        logger.debug(cmd)
        stdout = subprocess.check_output(cmd,shell=True,
                                         stderr=subprocess.STDOUT)
        logger.debug(stdout)

        cmd = 'unzip %s -d %s'%(tmpfile,tmpdir)
        logger.debug(cmd)
        stdout = subprocess.check_output(cmd,shell=True,
                                         stderr=subprocess.STDOUT)
        logger.debug(stdout)

        logger.debug("Creating %s..."%outfile)
        shutil.move(tmpfile.replace('.zip','.cmd'),outfile)
        os.remove(tmpfile)

        return outfile

    @classmethod
    def verify(cls, filename, survey, age, metallicity):
        age = age*1e9
        nlines=14
        with open(filename,'r') as f:
            lines = [f.readline() for i in range(nlines)]
            if len(lines) < nlines:
                msg = "Incorrect file size"
                raise Exception(msg)
                
            try:
                s = lines[2].split()[-2]
                assert dict_output[survey][:4] in s
            except:
                msg = "Incorrect survey:\n"+lines[2]
                raise Exception(msg)

            try:
                z = lines[5].split()[2]
                assert np.allclose(metallicity,float(z),atol=1e-3)
            except:
                msg = "Metallicity does not match:\n"+lines[5]
                raise Exception(msg)

            try:
                a = lines[13].split()[1]
                assert np.allclose(age,float(a),atol=1e-5)
            except:
                msg = "Age does not match:\n"+lines[13]
                raise Exception(msg)



#!/usr/bin/env python
"""
Analyze a simulated data set.
"""
__author__ = "Alex Drlica-Wagner"
import copy
import os
import time
import resource, psutil
from collections import OrderedDict as odict

import numpy
import numpy as np
import scipy.interpolate
import healpy
import numpy.lib.recfunctions as recfuncs
import fitsio

import ugali.observation.catalog
import ugali.observation.mask
import ugali.observation.roi
import ugali.utils.projector
import ugali.utils.stats
import ugali.analysis.scan

from ugali.utils.projector import gal2cel, cel2gal, sr2deg, mod2dist
from ugali.utils.healpix import ang2pix, pix2ang
from ugali.utils.logger import logger
from ugali.utils.config import Config
from ugali.utils import mlab

# Analysis flags
FLAGS = odict([])
FLAGS['FLAG_PROC'  ] = FLAG_PROC   = 0  # Simulation was processed
FLAGS['FLAG_NOPROC'] = FLAG_NOPROC = 1  # No processing
FLAGS['FLAG_NOBJ'  ] = FLAG_NOBJ   = 2  # Too many catalog objects
FLAGS['FLAG_FIT'   ] = FLAG_FIT    = 4  # Fit failure
FLAGS['FLAG_EBV'   ] = FLAG_EBV    = 8  # EBV value too large
FLAGS['FLAG_MEM'   ] = FLAG_MEM    = 16 # Memory error

def update_header_flags(filename):
    fits = fitsio.FITS(filename,'rw')
    for k,v in FLAGS.items():
        fits[1].write_key(k,v)

# Results dtypes
# FIXME: KERNEL should be removed for next run
DTYPES=[('TS','>f4'),
        ('FIT_KERNEL','S18'),('FIT_EXTENSION','>f4'),
        ('FIT_MASS','>f4'),('FIT_MASS_ERR','>f4'),
        ('FIT_DISTANCE','>f4'),('FIT_DISTANCE_ERR','>f4'),('FLAG','>i4'),
        ('RUNTIME','>f4'),('MEMORY','>i8')]

KB = 1024**1
MB = 1024**2
GB = 1024**3

class Analyzer(object):
    """
    Class for analyzing simulated data
    """
    def __init__(self, config, catfile=None, popfile=None):
        self.config = Config(config)
        self.population = self.read_population(popfile)
        self.catalog = self.read_catalog(catfile)
        self.mlimit = -1

    def get_memory_usage(self):
        """Get the memory usage of this process.

        Parameters
        ----------
        """
        process = psutil.Process()
        mem = process.memory_info()[0]
        return mem
    
        ## https://stackoverflow.com/a/7669482/4075339
        ## peak memory usage (kilobytes on Linux)
        #usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
        #return usage
        
    def get_memory_limit(self):
        """Get the hard memory limit from LSF.

        Parameters
        ----------
        None
        
        Returns
        -------
        mlimit : memory limit (bytes)
        """
        rsrc = resource.RLIMIT_AS
        soft, hard = resource.getrlimit(rsrc)
        if os.getenv('LSB_CG_MEMLIMIT') and os.getenv('LSB_HOSTS'):
            # Memory limit per core
            memlimit = int(os.getenv('LSB_CG_MEMLIMIT'), 16)
            # Number of cores
            ncores = len(os.getenv('LSB_HOSTS').split())
            #soft = ncores * memlimit - 100*1024**2
            soft = ncores * memlimit - GB
        return soft

    def set_memory_limit(self, mlimit):
        """Set the (soft) memory limit through setrlimit.
        
        Parameters
        ----------
        mlimit : soft memory limit (bytes)
        
        Returns
        -------
        soft, hard : memory limits (bytes)
        """
        rsrc = resource.RLIMIT_AS
        resource.setrlimit(rsrc, (mlimit, mlimit))
        self.mlimit = mlimit
        return resource.getrlimit(rsrc)

        #rsrc = resource.RLIMIT_AS
        #soft, hard = resource.getrlimit(rsrc)
        #resource.setrlimit(rsrc, (mlimit, hard))
        #self.mlimit, hard = resource.getrlimit(rsrc)
        #return (self.mlimit,hard)


    def read_population(self, filename=None):
        if not filename:
            filename = os.path.join(self.config['simulate']['dirname'],self.config['simulate']['popfile'])
        logger.info("Reading population file: %s"%filename)
        pop = ugali.utils.fileio.read(filename)
        pop.dtype.names = list(map(str.upper,pop.dtype.names))
        return pop

    def read_catalog(self, filename=None):
        if not filename:
            filename = os.path.join(self.config['simulate']['dirname'],self.config['simulate']['catfile'])
        logger.info("Reading catalog file: %s"%filename)
        catalog =  ugali.observation.catalog.Catalog(self.config,filenames=filename)
        catalog.data = mlab.rec_append_fields(catalog.data,
                                           names=['PIX8','PIX4096'],
                                           arrs=np.zeros((2,len(catalog.lon)),dtype='>i8'))
        return catalog

    def read_results(self, filename):
        logger.info("Reading results file: %s"%filename)
        results = ugali.utils.fileio.read(filename)
        return results
        
    def create_results(self, population=None):
        """ Create the results array of len(population)

        Parameters:
        -----------
        population : population array (self.population if None)
        
        Returns:
        --------
        results : array of len(population)
        """
        if population is None: population = self.population
        size = len(population)
        results = np.array(np.nan*np.ones(size),dtype=DTYPES)
        results = recfuncs.merge_arrays([population,results],
                                             flatten=True,
                                             asrecarray=False,usemask=False)
        results['TS'] = -np.inf
        results['FLAG'] = FLAG_NOPROC
        results['MEMORY'] = -1

        return results

    def write_results(self, filename,**kwargs):
        """ Write results array to a file.

        Parameters:
        -----------
        filename : output file name
        kwargs   : arguments passed to fileio.write
        
        Returns:
        --------
        None
        """
        ugali.utils.fileio.write(filename,self.results,**kwargs)
        update_header_flags(filename)

    def runall(self,outfile=None,mc_source_id=None,rerun=False):
        """Run all sources in population.
        
        Parameters:
        -----------
        outfile      : file to write output to
        mc_source_id : list of sources to process (None is all sources)

        Returns:
        --------
        results      : processing results
        """
        if mc_source_id is None:
            mc_source_id = np.unique(self.catalog.mc_source_id)

        # Select only systems that are in the catalog
        sel = np.in1d(self.population['MC_SOURCE_ID'],mc_source_id)
        
        if not sel.sum():
            msg = "Requested MC_SOURCE_IDs not found in population."
            raise ValueError(msg)

        if os.path.exists(outfile) and rerun:
            # read the results from the existing outfile
            self.results = self.read_results(outfile)
        else:
            # create the results
            self.results = self.create_results(population=self.population[sel])
        
        if not np.in1d(mc_source_id,self.results['MC_SOURCE_ID']).all():
            msg = "Requested MC_SOURCE_IDs not found in results."
            raise ValueError(msg)
    
        if outfile: 
            logger.info("Writing %s..."%outfile)
            self.write_results(outfile,clobber=True)

        for i,r in enumerate(self.results):
            # Skip if not in mc_source_id list
            if self.results[i]['MC_SOURCE_ID'] not in mc_source_id:
                msg = "%i skipped."%self.results[i]['MC_SOURCE_ID']
                logger.info(msg)
                continue

            # Rerun in NOPROC or MEM flagged
            if (self.results[i]['FLAG'] & (FLAG_NOPROC | FLAG_MEM)) == 0:
                msg = "%i already processed."%self.results[i]['MC_SOURCE_ID']
                logger.info(msg)
                continue

            start_time = time.time()

            try: 
                self.runone(i)
            except MemoryError as e:
                msg = "Memory usage exceeded %.3f GB"%(self.mlimit/GB)
                logger.warn(msg)
                self.results[i]['FLAG'] |= FLAG_MEM
            except Exception as e:
                logger.error(str(e))
                self.results[i]['FLAG'] |= FLAG_FIT
                
            runtime = time.time() - start_time
            self.results[i]['MEMORY'] = self.get_memory_usage()
            self.results[i]['RUNTIME'] = runtime

            logger.info("Fit parameter values:")
            for d in DTYPES:
                logger.info('\t%s: %s'%(d[0], self.results[i][d[0]]))

            logger.info("Memory usage: %.3f GB"%(self.get_memory_usage()/GB))

            if (i%self.config['simulate']['save'])==0 and outfile:
                logger.info("Writing %s..."%outfile)
                self.write_results(outfile,clobber=True)

        if outfile: 
            logger.info("Writing %s..."%outfile)
            self.write_results(outfile,clobber=True)

        return self.results

    #from memory_profiler import profile
    #@profile
    def runone(self, i):
        """ Run one simulation.

        Parameters:
        -----------
        i : index of the simulation to run
        
        Returns:
        --------
        results : result array
        """
        results = self.results
        results[i]['FLAG'] = FLAG_PROC

        params = dict(list(zip(results[i].dtype.names,results[i])))
        size = len(results)
        lon,lat = params['RA'],params['DEC']
        distance_modulus = params['DISTANCE_MODULUS']
        mc_source_id = params['MC_SOURCE_ID']
        extension=np.degrees(np.arctan(params['R_PHYSICAL']/params['DISTANCE']))

        logger.info('\n(%i/%i); (id, lon, lat, mod, ext) = (%i, %.2f, %.2f, %.1f, %.3f)'%(i+1,size,mc_source_id,lon,lat,distance_modulus,extension))
        
        if params['EBV'] > 0.2:
            msg = "High reddening region; skipping..."
            logger.warn(msg)
            results[i]['FLAG'] |= FLAG_EBV
            #raise Exception(msg)
            #results[i]['TS'] = np.nan
            #return

        # This links to the parameters in the data scan
        section = 'scan'
        # This links to the parameters in the simulate section
        #section = 'simulate'

        source=ugali.analysis.loglike.createSource(self.config,section=section,lon=lon,lat=lat)
        
        logger.info("Reading data catalog...")
        obs = ugali.analysis.loglike.createObservation(self.config,lon=lon,lat=lat)
            

        # Select just the simulated target of interest
        logger.info("Merging simulated catalog...")
        data = self.catalog.data[self.catalog.mc_source_id == mc_source_id].copy()
        data = np.array(data[list(obs.catalog.data.dtype.names)],
                        dtype=obs.catalog.data.dtype)
        obs.catalog = ugali.observation.catalog.Catalog(self.config, data=np.concatenate([obs.catalog.data,data]))

        loglike = ugali.analysis.loglike.LogLikelihood(self.config,obs,source)

        # Mitigate memory overflow issues by cutting objects with
        # too many catalog stars
        if len(loglike.catalog) > 5e5: # 1e5
            msg = "Large catalog (N_CATALOG = %i)."%len(loglike.catalog)
            logger.warn(msg)
            results[i]['FLAG'] |= FLAG_NOBJ

        grid = ugali.analysis.scan.GridSearch(self.config,loglike)
        self.grid = grid
        self.loglike = self.grid.loglike

        pix = self.loglike.roi.indexTarget(lon,lat)

        # ADW: Should fit_distance be free in order to model search procedure?
        if self.config['simulate'].get('fit_distance',False):
            fit_distance = None
        else:
            idx = np.fabs(grid.distance_modulus_array-distance_modulus).argmin()
            fit_distance = grid.distance_modulus_array[idx]

        try:
            grid.search(coords=(lon,lat),distance_modulus=fit_distance)
            results[i]['FLAG'] &= ~FLAG_FIT
        except ValueError as e:
            logger.error(str(e))
            results[i]['FLAG'] |= FLAG_FIT

        mle = grid.mle()
        
        distance_idx = np.fabs(grid.distance_modulus_array-mle['distance_modulus']).argmin()

        ts = 2*grid.loglike_array[distance_idx][pix]

        results[i]['FIT_KERNEL'] = grid.loglike.kernel.name
        results[i]['TS'] = ts
        results[i]['FIT_MASS'] = grid.stellar_mass_conversion*mle['richness']
        results[i]['FIT_DISTANCE'] = mle['distance_modulus']
        results[i]['FIT_EXTENSION'] = grid.loglike.kernel.extension
        
        err = grid.err()
        richness_err = (err['richness'][1]-err['richness'][0])/2.
        results[i]['FIT_MASS_ERR'] = grid.stellar_mass_conversion*richness_err
             
        distance_modulus_err = (err['distance_modulus'][1]-err['distance_modulus'][0])/2.
        results[i]['FIT_DISTANCE_ERR'] = distance_modulus_err

        """
        fit_extension = [0.02, 0.07, 0.15] # half light radii (deg)

        # ADW: This won't work since we add extension inside the search loop
        for ext in fit_extension:
            grid.loglike.set_params(extension = ext)
            # Fit failures are often due to fracdet = 0
            try:
                grid.search(coords=(lon,lat),distance_modulus=fit_distance)
                results[i]['FLAG'] &= ~FLAG_FIT
            except ValueError as e:
                logger.error(str(e))
                results[i]['FLAG'] |= FLAG_FIT
                continue
            mle = grid.mle()
            
            ts = 2*grid.loglike_array[distance_idx][pix]
            if ts <= results[i]['TS']: 
                logger.info("No TS increase; continuing...")
                continue

            results[i]['FIT_KERNEL'] = grid.loglike.kernel.name
            results[i]['TS'] = ts
            results[i]['FIT_MASS'] = grid.stellar_mass_conversion*mle['richness']
            results[i]['FIT_DISTANCE'] = fit_distance #mle['distance_modulus']
            results[i]['FIT_EXTENSION'] = ext

            err = grid.err()
            richness_err = (err['richness'][1]-err['richness'][0])/2.
            results[i]['FIT_MASS_ERR'] = grid.stellar_mass_conversion*richness_err
             
            distance_modulus_err = (err['distance_modulus'][1]-err['distance_modulus'][0])/2.
            results[i]['FIT_DISTANCE_ERR'] = distance_modulus_err
            """
        return results

    run = runall

if __name__ == "__main__":
    import ugali.utils.parser
    parser = ugali.utils.parser.Parser(description=__doc__)
    parser.add_config()
    parser.add_argument('-p','--popfile',default=None,
                        help='simulated population input file')
    parser.add_argument('-c','--catfile',default=None,
                        help='simulated catalog input file')
    parser.add_argument('-o','--outfile',default=None,
                        help='output results file')
    parser.add_argument('-i','--mc-source-id',default=None,type=int,action='append',
                        help='specific source id to run')
    parser.add_argument('-m','--mlimit',default=None,type=int,
                        help='limit memory usage')
    parser.add_argument('-r','--rerun',action='store_true',
                        help='rerun failed jobs')

    #parser.add_force()
    #parser.add_debug()
    parser.add_verbose()
    args = parser.parse_args()

    analyzer = Analyzer(args.config,args.catfile,args.popfile)

    if args.mlimit is not None:
        if args.mlimit == 0:
            mlimit = analyzer.get_memory_limit()
        else:
            mlimit = args.mlimit * GB
        soft,hard = analyzer.set_memory_limit(mlimit)
        logger.info("Setting memory limit to %.3f GB"%(soft/GB))

    if args.mc_source_id is None:
        basename = os.path.splitext(args.catfile)[0]
        imin,imax = list(map(int,basename.rsplit('_',1)[-1].split('-')))
        args.mc_source_id = np.arange(imin,imax+1)

    analyzer.run(outfile=args.outfile,mc_source_id=args.mc_source_id,
                 rerun=args.rerun)



#!/usr/bin/env python
"""
Create and run an individual likelihood analysis.
"""
import os
import sys
from collections import OrderedDict as odict

import numpy
import numpy as np
import healpy
import fitsio

import ugali.utils.skymap
import ugali.analysis.loglike
from ugali.analysis.loglike import LogLikelihood, createSource, createObservation
from ugali.analysis.source import Source
from ugali.utils.parabola import Parabola

from ugali.utils.batch import LSF
from ugali.utils.config import Config
from ugali.utils.logger import logger
from ugali.utils.healpix import superpixel, subpixel, pix2ang, ang2pix
from ugali.utils.healpix import write_partial_map

############################################################

class Scan(object):
    """
    The base of a likelihood analysis scan.

    FIXME: This really does nothing now and should be absorbed into
    GridSearch (confusing now).
    """
    def __init__(self, config, coords):
        self.config = Config(config)
        # Should only be one coordinate
        if len(coords) != 1: raise Exception('Must specify one coordinate.')
        self.lon,self.lat,radius = coords[0]
        self._setup()
        

    def _setup(self):
        self.observation = createObservation(self.config,self.lon,self.lat)
        self.source = createSource(self.config,'scan',lon=self.lon,lat=self.lat)
        loglike=LogLikelihood(self.config,self.observation,self.source)
        self.grid = GridSearch(self.config,loglike)

    @property
    def loglike(self):
        return self.grid.loglike

    def run(self, coords=None, debug=False):
        """
        Run the likelihood grid search
        """
        #self.grid.precompute()
        self.grid.search(coords=coords)
        return self.grid
        
    def write(self, outfile):
        self.grid.write(outfile)

def createGridSearch(config,lon,lat):
    config = Config(config)
    obs = createObservation(config,lon,lat)
    src = createSource(config,'scan',lon=lon,lat=lat)
    loglike=LogLikelihood(config,obs,src)
    return GridSearch(config,loglike)
    

############################################################

class GridSearch:

    def __init__(self, config, loglike): # What it should be...
        """
        Object to efficiently search over a grid of ROI positions.

        Parameters:
        -----------
        config  : Configuration object or filename.
        loglike : Log-likelihood object

        Returns:
        --------
        grid    : GridSearch instance
        """

        self.config = Config(config)
        self.loglike = loglike
        self.source = self.loglike.source
        self.roi  = self.loglike.roi
        self.mask = self.loglike.mask

        logger.info(str(self.loglike))

        self.stellar_mass_conversion = self.source.stellar_mass()
        self.distance_modulus_array = np.asarray(self.config['scan']['distance_modulus_array'])
        self.extension_array = np.asarray(self.config['scan'].get('extension_array',[self.source.extension]))


    def search(self, coords=None, distance_modulus=None, extension=None, tolerance=1.e-2):
        """
        Organize a grid search over ROI target pixels, distance
        moduli, and extensions. If coords, distance_modulus, or
        extension is specified, then the nearest value in the
        predefined scan grid is used. ***This may be different than
        the input value.**

        Parameters
        ----------
        coords : (float,float)
            coordinate to search (matched to nearest scan value)
        distance_modulus : float
            distance modulus to search (matched to nearest scan value)
        extension : float
            extension to search (matched to nearest scan value)
        tolerance : float
            tolerance on richness maximization

        Returns
        -------
        None 
        """
        nmoduli = len(self.distance_modulus_array)
        npixels = len(self.roi.pixels_target)
        self.loglike_array              = np.zeros([nmoduli,npixels],dtype='f4')
        self.richness_array             = np.zeros([nmoduli,npixels],dtype='f4')
        self.stellar_mass_array         = np.zeros([nmoduli,npixels],dtype='f4')
        self.fraction_observable_array  = np.zeros([nmoduli,npixels],dtype='f4')
        self.extension_fit_array        = np.zeros([nmoduli,npixels],dtype='f4')
        if self.config['scan']['full_pdf']:
            # DEPRECATED: ADW 2019-04-27
            DeprecationWarning("'full_pdf' is deprecated.")
            self.richness_lower_array       = np.zeros([nmoduli,npixels],dtype='f4')
            self.richness_upper_array       = np.zeros([nmoduli,npixels],dtype='f4')
            self.richness_ulimit_array      = np.zeros([nmoduli,npixels],dtype='f4')

        # Specific pixel/distance_modulus
        coord_idx, distance_modulus_idx, extension_idx = None, None, None
        if coords is not None:
            # Match to nearest grid coordinate index
            coord_idx = self.roi.indexTarget(coords[0],coords[1])
        if distance_modulus is not None:
            # Match to nearest distance modulus index
            distance_modulus_idx=np.fabs(self.distance_modulus_array-distance_modulus).argmin()
        if extension is not None:
            # Match to nearest extension
            extension_idx=np.fabs(self.extension_array-extension).argmin()

        lon, lat = self.roi.pixels_target.lon, self.roi.pixels_target.lat
            
        logger.info('Looping over distance moduli in grid search ...')
        for ii, distance_modulus in enumerate(self.distance_modulus_array):
            # Specific distance
            if distance_modulus_idx is not None:
                if ii != distance_modulus_idx: continue

            logger.info('  (%-2i/%i) Distance Modulus=%.1f ...'%(ii+1,nmoduli,distance_modulus))

            # No objects, continue
            if len(self.loglike.catalog) == 0: 
                logger.warn("No catalog objects")
                continue

            # Set distance_modulus once to save time
            self.loglike.set_params(distance_modulus=distance_modulus)
            # Loop over pixels
            for jj in range(0, npixels):
                # Specific pixel
                if coord_idx is not None:
                    if jj != coord_idx: continue

                # Set kernel location
                self.loglike.set_params(lon=lon[jj],lat=lat[jj])

                loglike = 0
                # Loop over extensions
                for kk,ext in enumerate(self.extension_array):
                    # Specific extension
                    if extension_idx is not None:
                        if kk != extension_idx: continue

                    # Set extension
                    self.loglike.set_params(extension=ext)

                    # Doesn't re-sync distance_modulus each time
                    self.loglike.sync_params()

                    # Maximize the likelihood with respect to richness
                    loglike,rich,p = self.loglike.fit_richness()

                    if loglike < self.loglike_array[ii][jj]:
                        # No loglike increase, continue
                        continue

                    self.loglike_array[ii][jj],self.richness_array[ii][jj], parabola = loglike,rich,p
                    self.stellar_mass_array[ii][jj] = self.stellar_mass_conversion*self.richness_array[ii][jj]
                    self.fraction_observable_array[ii][jj] = self.loglike.f
                    self.extension_fit_array[ii][jj] = self.source.extension

                # ADW: Careful, we are leaving the extension at the
                # last value in the array, not at the maximum...

                # Debug output
                args = (jj+1, npixels, lon[jj], lat[jj],
                        2.*self.loglike_array[ii][jj], 
                        self.stellar_mass_array[ii][jj],
                        self.fraction_observable_array[ii][jj],
                        self.extension_fit_array[ii][jj]
                        )
                msg  = '    (%-3i/%i) Candidate at (%.2f, %.2f) ... '
                msg += 'TS=%.1f, Mstar=%.2g, ObsFrac=%.2g, Ext=%.2g'
                logger.debug(msg%args)

                """
                # This is debugging output
                if self.config['scan']['full_pdf']:
                    DeprecationWarning("'full_pdf' is deprecated.")
                    self.richness_lower_array[ii][jj], self.richness_upper_array[ii][jj] = self.loglike.richness_interval(0.6827)
                    
                    self.richness_ulimit_array[ii][jj] = parabola.bayesianUpperLimit(0.95)

                    args = (
                        2. * self.loglike_array[ii][jj],
                        self.stellar_mass_conversion*self.richness_array[ii][jj],
                        self.stellar_mass_conversion*self.richness_lower_array[ii][jj],
                        self.stellar_mass_conversion*self.richness_upper_array[ii][jj],
                        self.stellar_mass_conversion*self.richness_ulimit_array[ii][jj]
                    )
                    msg = 'TS=%.1f, Stellar Mass=%.1f (%.1f -- %.1f @ 0.68 CL, < %.1f @ 0.95 CL)'%(args)
                    logger.debug(msg)
                """
                
            jj_max = self.loglike_array[ii].argmax()
            args = (
                jj_max+1, npixels, lon[jj_max], lat[jj_max],
                2. * self.loglike_array[ii][jj_max], 
                self.stellar_mass_conversion * self.richness_array[ii][jj_max],
                self.extension_fit_array[ii][jj_max]
            )
            msg = '  (%-3i/%i) Max at (%.2f, %.2f) : TS=%.1f, Mstar=%.2g, Ext=%.2f'%(args)
            logger.info(msg)

    def mle(self):
        a = self.loglike_array
        j,k = np.unravel_index(a.argmax(),a.shape)
        mle = odict()
        mle['richness'] = self.richness_array[j][k]
        mle['lon'] = self.roi.pixels_target.lon[k]
        mle['lat'] = self.roi.pixels_target.lat[k]
        mle['distance_modulus'] = self.distance_modulus_array[j]
        mle['extension'] = float(self.source.extension)
        mle['ellipticity'] = float(self.source.ellipticity)
        mle['position_angle'] = float(self.source.position_angle)
        # ADW: FIXME!
        try: 
            mle['age'] = np.mean(self.source.age)
            mle['metallicity'] = np.mean(self.source.metallicity)
        except AttributeError:
            mle['age'] = np.nan
            mle['metallicity'] = np.nan
            
        return mle

    def err(self):
        """
        A few rough approximations of the fit uncertainty. These
        values shouldn't be trusted for anything real (use MCMC instead).
        """
        # Initiallize error to nan
        err = odict(self.mle())
        err.update([(k,np.nan*np.ones(2)) for k in err.keys()])

        # Find the maximum likelihood
        a = self.loglike_array
        j,k = np.unravel_index(a.argmax(),a.shape)

        self.loglike.set_params(distance_modulus=self.distance_modulus_array[j],
                                lon=self.roi.pixels_target.lon[k],
                                lat=self.roi.pixels_target.lat[k])
        self.loglike.sync_params()

        # Find the error in richness, starting at maximum
        lo,hi = np.array(self.loglike.richness_interval())
        err['richness'] = np.array([lo,hi])

        # ADW: This is a rough estimate of the distance uncertainty 
        # hacked to keep the maximum distance modulus on a grid index

        # This is a hack to get the confidence interval to play nice...
        # Require at least three points.
        if (a[:,k]>0).sum() >= 3:
            parabola = Parabola(np.insert(self.distance_modulus_array,0,0.), 
                                np.insert(a[:,k],0,0.) )
            lo,hi = np.array(parabola.confidenceInterval())
            err['distance_modulus'] = self.distance_modulus_array[j] + (hi-lo)/2.*np.array([-1.,1.])

        # ADW: Could estimate lon and lat from the grid.
        # This is just concept right now...
        if (a[j,:]>0).sum() >= 10:
            delta_ts = 2*(a[j,k] - a[j,:])
            pix = np.where(a[j,:][delta_ts < 2.71])[0]
            lons = self.roi.pixels_target.lon[pix]
            lats = self.roi.pixels_target.lat[pix]
            err['lon'] = np.array([ np.min(lons),np.max(lons)])
            err['lat'] = np.array([ np.min(lats),np.max(lats)])

        return err

    def write(self, outfile):
        """
        Save the likelihood results as a sparse HEALPix map.
        """
        data = odict()
        data['PIXEL']=self.roi.pixels_target
        # Full data output (too large for survey)
        if self.config['scan']['full_pdf']:
            DeprecationWarning("'full_pdf' is deprecated.")
            data['LOG_LIKELIHOOD']=self.loglike_array.T
            data['RICHNESS']=self.richness_array.T
            data['RICHNESS_LOWER']=self.richness_lower_array.T
            data['RICHNESS_UPPER']=self.richness_upper_array.T
            data['RICHNESS_LIMIT']=self.richness_ulimit_array.T
            #data['STELLAR_MASS']=self.stellar_mass_array.T
            data['FRACTION_OBSERVABLE']=self.fraction_observable_array.T
        else:
            data['LOG_LIKELIHOOD']=self.loglike_array.T
            data['RICHNESS']=self.richness_array.T
            data['FRACTION_OBSERVABLE']=self.fraction_observable_array.T

        # Convert to 32bit float
        for k in list(data.keys())[1:]:
            data[k] = data[k].astype('f4',copy=False)
            
        # Stellar mass can be calculated from STELLAR * RICHNESS
        header = odict()
        header['STELLAR']=round(self.stellar_mass_conversion,8)
        header['LKDNSIDE']=self.config['coords']['nside_likelihood']
        header['LKDPIX']=ang2pix(self.config['coords']['nside_likelihood'],
                                 self.roi.lon,self.roi.lat)
        header['NROI']=self.roi.inROI(self.loglike.catalog_roi.lon,
                                      self.loglike.catalog_roi.lat).sum()
        header['NANNULUS']=self.roi.inAnnulus(self.loglike.catalog_roi.lon,
                                              self.loglike.catalog_roi.lat).sum()
        header['NINSIDE']=self.roi.inInterior(self.loglike.catalog_roi.lon,
                                              self.loglike.catalog_roi.lat).sum()
        header['NTARGET']=self.roi.inTarget(self.loglike.catalog_roi.lon,
                                            self.loglike.catalog_roi.lat).sum()

        # Flatten if there is only a single distance modulus
        # ADW: Is this really what we want to do?
        if len(self.distance_modulus_array) == 1:
            for key in data:
                data[key] = data[key].flatten()

        logger.info("Writing %s..."%outfile)
        write_partial_map(outfile,data,
                          nside=self.config['coords']['nside_pixel'],
                          header=header,
                          clobber=True
                          )
        
        fitsio.write(outfile,
                     dict(DISTANCE_MODULUS=self.distance_modulus_array.astype('f4',copy=False)),
                     extname='DISTANCE_MODULUS',
                     clobber=False)

if __name__ == "__main__":
    import ugali.utils.parser
    description = "Script for executing the likelihood scan."
    parser = ugali.utils.parser.Parser(description=description)
    parser.add_config()
    parser.add_argument('outfile',metavar='outfile.fits',help='Output fits file.')
    parser.add_argument('-m','--mlimit',nargs='?',default=-1, type=int,
                        help='Memory limit (KB)')
    parser.add_debug()
    parser.add_verbose()
    parser.add_coords(required=True,radius=False)
    opts = parser.parse_args()

    if len(opts.coords) != 1: 
        raise Exception('Must specify exactly one coordinate.')
    lon,lat,radius = opts.coords[0]

    #if opts.mlimit > 0:
    if True:
        # Convert from KB to GB
        mlimit,_ = LSF.get_memory_limit()
        logger.info("Setting memory limit: %.1fGB"%(mlimit/1024.**3))
        soft,hard = LSF.set_memory_limit(mlimit)
        logger.info("Memory limit: %.1fGB"%(soft/1024.**3))

    grid = createGridSearch(opts.config,lon,lat)
    if not opts.debug:
        result = grid.search()
        grid.write(opts.outfile)



#!/usr/bin/env python
"""
Python utilities.
"""
import six
import numpy as np

def iterable(obj):
    """return true if *obj* is iterable"""
    try:
        iter(obj)
    except TypeError:
        return False
    return True

def isstring(obj):
    """Python 2/3 compatible string check"""
    return isinstance(obj, six.string_types)

def rec_append_fields(rec, names, arrs, dtypes=None):
    """
    Return a new record array with field names populated with data
    from arrays in *arrs*.  If appending a single field, then *names*,
    *arrs* and *dtypes* do not have to be lists. They can just be the
    values themselves.
    """
    if (not isstring(names) and iterable(names) and len(names) and isstring(names[0])):
        if len(names) != len(arrs):
            raise ValueError("number of arrays do not match number of names")
    else:  # we have only 1 name and 1 array
        names = [names]
        arrs = [arrs]
    arrs = list(map(np.asarray, arrs))
    if dtypes is None:
        dtypes = [a.dtype for a in arrs]
    elif not iterable(dtypes):
        dtypes = [dtypes]
    if len(arrs) != len(dtypes):
        if len(dtypes) == 1:
            dtypes = dtypes * len(arrs)
        else:
            raise ValueError("dtypes must be None, a single dtype or a list")
    old_dtypes = rec.dtype.descr
    if six.PY2:
        old_dtypes = [(name.encode('utf-8'), dt) for name, dt in old_dtypes]
    newdtype = np.dtype(old_dtypes + list(zip(names, dtypes)))
    newrec = np.recarray(rec.shape, dtype=newdtype)
    for field in rec.dtype.fields:
        newrec[field] = rec[field]
    for name, arr in zip(names, arrs):
        newrec[name] = arr
    return newrec



#!/usr/bin/env python
"""
For pixelizing catalogs and masks.
"""

#FIXME: ADW This needs to be updated to use fitsio

import os
from os.path import join
import glob
import collections

import fitsio
import numpy as np
import numpy.lib.recfunctions as recfuncs
import healpy as hp

#import ugali.utils.binning
#import ugali.utils.skymap
from ugali.utils.projector import cel2gal, gal2cel
from ugali.utils import healpix, mlab
from ugali.utils.healpix import ang2pix, pix2ang, superpixel

from ugali.utils.shell import mkdir
from ugali.utils.logger import logger
from ugali.utils.config import Config
import ugali.utils.fileio

def pixelizeCatalog(infiles, config, force=False):
    """
    Break catalog into chunks by healpix pixel.
    
    Parameters:
    -----------
    infiles : List of input files
    config  : Configuration file
    force   : Overwrite existing files (depricated)
    
    Returns:
    --------
    None
    """
    nside_catalog = config['coords']['nside_catalog']
    nside_pixel = config['coords']['nside_pixel']
    coordsys = config['coords']['coordsys'].upper()
    outdir = mkdir(config['catalog']['dirname'])
    filenames = config.getFilenames()
    lon_field = config['catalog']['lon_field'].upper()
    lat_field = config['catalog']['lat_field'].upper()

    # ADW: It would probably be better (and more efficient) to do the
    # pixelizing and the new column insertion separately.
    for i,filename in enumerate(infiles):
        logger.info('(%i/%i) %s'%(i+1, len(infiles), filename))
        data = fitsio.read(filename)
        logger.info("%i objects found"%len(data))
        if not len(data): continue

        columns = map(str.upper,data.dtype.names)
        names,arrs = [],[]

        if (lon_field in columns) and (lat_field in columns):
            lon,lat = data[lon_field],data[lat_field]
        elif coordsys == 'GAL':
            msg = "Columns '%s' and '%s' not found."%(lon_field,lat_field)
            msg += "\nConverting from RA,DEC"
            logger.warning(msg)
            lon,lat = cel2gal(data['RA'],data['DEC'])
            names += [lon_field,lat_field]
            arrs  += [lon,lat]
        elif coordsys == 'CEL':
            msg = "Columns '%s' and '%s' not found."%(lon_field,lat_field)
            msg += "\nConverting from GLON,GLAT"
            lon,lat = gal2cel(data['GLON'],data['GLAT'])
            names  += [lon_field,lat_field]
            arrs   += [lon,lat]

        cat_pix = ang2pix(nside_catalog,lon,lat)
        pix_pix = ang2pix(nside_pixel,lon,lat)
        cat_pix_name = 'PIX%i'%nside_catalog
        pix_pix_name = 'PIX%i'%nside_pixel

        try:
            names += [cat_pix_name,pix_pix_name]
            arrs  += [cat_pix,pix_pix]
            data=mlab.rec_append_fields(data,names=names,arrs=arrs)
        except ValueError as e:
            logger.warn(str(e)+'; not adding column.')
            #data[cat_pix_name] = cat_pix
            #data[pix_pix_name] = pix_pix
                               
        for pix in np.unique(cat_pix):
            logger.debug("Processing pixel %s"%pix)

            arr = data[cat_pix == pix]
            outfile = filenames.data['catalog'][pix]

            if not os.path.exists(outfile):
                logger.debug("Creating %s"%outfile)
                out=fitsio.FITS(outfile,mode='rw')
                out.write(arr)

                hdr=healpix.header_odict(nside=nside_catalog,
                                                     coord=coordsys[0])
                for key in ['PIXTYPE','ORDERING','NSIDE','COORDSYS']:
                    out[1].write_key(*list(hdr[key].values()))
                out[1].write_key('PIX',pix,comment='HEALPIX pixel for this file')
            else:
                out=fitsio.FITS(outfile,mode='rw')
                out[1].append(arr)

            logger.debug("Writing %s"%outfile)
            out.close()

def pixelizeDensity(config, nside=None, force=False):
    if nside is None: 
        nside = config['coords']['nside_likelihood']
    coordsys = config['coords']['coordsys'].upper()
    filenames = config.getFilenames()
    infiles = filenames[~filenames['catalog'].mask]
    lon_field = config['catalog']['lon_field'].upper()
    lat_field = config['catalog']['lat_field'].upper()

    for ii,f in enumerate(infiles.data):
        infile = f['catalog']
        pix = f['pix']
        logger.info('(%i/%i) %s'%(ii+1, len(infiles), infile))

        outfile = config['data']['density']%pix
        if os.path.exists(outfile) and not force: 
            logger.info("Found %s; skipping..."%outfile)
            continue
            
        outdir = mkdir(os.path.dirname(outfile))
        pixels, density = stellarDensity(infile,nside,
                                         lon_field=lon_field,lat_field=lat_field)

        data = dict(PIXEL=pixels,DENSITY=density)
        healpix.write_partial_map(outfile,data,nside=nside,coord=coordsys[0])


def stellarDensity(infile, nside=256, lon_field='RA', lat_field='DEC'): 
    area = hp.nside2pixarea(nside,degrees=True)
    logger.debug("Reading %s"%infile)
    data = fitsio.read(infile,columns=[lon_field,lat_field])

    lon,lat = data[lon_field],data[lat_field]
    pix = ang2pix(nside,lon,lat)
    counts = collections.Counter(pix)
    pixels, number = np.array(sorted(counts.items())).T
    density = number/area

    return pixels, density

if __name__ == "__main__":
    from optparse import OptionParser
    usage = "Usage: %prog  [options] input"
    description = "python script"
    parser = OptionParser(usage=usage,description=description)
    (opts, args) = parser.parse_args()



"""
Tools for making maps of the sky with healpix. Used by simulations.
"""

import numpy as np
import healpy as hp

import ugali.utils.projector
from ugali.utils.healpix import superpixel,subpixel,ang2pix,pix2ang,query_disc
from ugali.utils.healpix import read_partial_map
from ugali.utils.logger import logger
from ugali.utils.config import Config

############################################################

def surveyPixel(lon, lat, nside_pix, nside_subpix = None):
    """
    Return the set of HEALPix pixels that cover the given coordinates at resolution nside.
    Optionally return the set of subpixels within those pixels at resolution nside_subpix
    """
    pix = np.unique(ang2pix(nside_pix, lon, lat))
    if nside_subpix is None:
        return pix
    else:
        subpix_array = []
        for ii in range(0, len(pix)):
            subpix = subpixel(pix[ii], nside_pix, nside_subpix)
            subpix_array.append(subpix)
        return pix, np.array(subpix_array)

def inFootprint(config, pixels, nside=None):
    """
    Open each valid filename for the set of pixels and determine the set 
    of subpixels with valid data.

    Parameters
    ----------
    config : config
        Configuration (file or object)
    pixels : array or int
        List of pixels to create footprint for
    nside  : int, optional
        Healpix nside
        
    Returns
    -------
    inside : array
        Boolean array of whether pixel is in footprint
    """
    logger.info("Calculating survey footprint...")

    config = Config(config)
    nside_catalog    = config['coords']['nside_catalog']
    nside_likelihood = config['coords']['nside_likelihood']
    nside_pixel      = config['coords']['nside_pixel']

    if np.isscalar(pixels): pixels = np.array([pixels])
    if nside is None: nside = nside_likelihood

    filenames = config.getFilenames()
    catalog_pixels = filenames['pix'].compressed()

    inside = np.zeros(len(pixels), dtype=bool)
    if not nside_catalog:
        catalog_pix = [0]
    else:
        catalog_pix = superpixel(pixels,nside,nside_catalog)
        catalog_pix = np.intersect1d(catalog_pix,catalog_pixels)

    fnames = filenames[catalog_pix]

    # Load the first mask
    logger.debug("Loading %s"%fnames['mask_1'])
    _nside,subpix1,val1 = read_partial_map(fnames['mask_1'],'MAGLIM',
                                           fullsky=False,multiproc=8)
    # Load the second mask
    logger.debug("Loading %s"%fnames['mask_2'])
    _nside,subpix2,val2 = read_partial_map(fnames['mask_2'],'MAGLIM',
                                           fullsky=False,multiproc=8)
    # Run the subpixels
    subpix = np.intersect1d(subpix1,subpix2)
    superpix = np.unique(superpixel(subpix,nside_pixel,nside))
    inside |= np.in1d(pixels, superpix)
    
    return inside

############################################################

def allSkyCoordinates(nside):
    """
    Generate a set of coordinates at the centers of pixels of resolutions nside across the full sky. 
    """
    lon,lat = pix2ang(nside, np.arange(0, hp.nside2npix(nside)))
    return lon, lat

def coarseFootprint(input, nside_pix):
    """
    Generate a coarse healpix mask of booleans from a finer healpix
    mask or a set of (lon, lat) coordinates.

    Parameters:
    -----------
    input :     (1) full HEALPix mask of booleans, or (2) a set of (lon, lat) coordinates for catalog objects that define the occupied pixels.
    nside_pix : nside_pix is meant to be at coarser (or equivalent) resolution than the input mask or catalog object positions
    so that gaps from star holes, bleed trails, cosmic rays, etc. are filled in. 

    Returns:
    --------
    lon,lat,area : Return the longitude and latitude of the random positions (deg) and the total area (deg^2).
    """
    input = np.array(input)
    if len(input.shape) == 1:
        if hp.npix2nside(len(input)) < nside_pix:
            logger.warning('Expected coarser resolution nside_pix in skymap.randomPositions')
        subpix = np.nonzero(input)[0] # All the valid pixels in the mask at the NSIDE for the input mask
        lon, lat = pix2ang(hp.npix2nside(len(input)), subpix)
    elif len(input.shape) == 2:
        lon, lat = input[0], input[1] # All catalog object positions
    else:
        logger.warning('Unexpected input dimensions for skymap.randomPositions')
    pix = surveyPixel(lon, lat, nside_pix)

    # Area with which the random points are thrown
    area = len(pix) * hp.nside2pixarea(nside_pix, degrees=True)

    # Create mask at the coarser resolution
    mask = np.tile(False, hp.nside2npix(nside_pix))
    mask[pix] = True
    
    return mask

def randomPositions(input, nside_pix, n=1):
    """
    Generate n random positions within a full HEALPix mask of booleans, or a set of (lon, lat) coordinates.

    Parameters:
    -----------
    input :     (1) full HEALPix mask of booleans, or (2) a set of (lon, lat) coordinates for catalog objects that define the occupied pixels.
    nside_pix : nside_pix is meant to be at coarser resolution than the input mask or catalog object positions
    so that gaps from star holes, bleed trails, cosmic rays, etc. are filled in. 
    n     : number of random points

    Returns:
    --------
    lon,lat,area : Return the longitude and latitude of the random positions (deg) and the total area (deg^2).

    """
    mask = coarseFootprint(input, nside_pix)
    area = mask.sum() * hp.nside2pixarea(nside_pix, degrees=True)

    # Estimate the number of points that need to be thrown based off
    # coverage fraction of the HEALPix mask
    coverage_fraction = float(np.sum(mask)) / len(mask) 
    n_throw = int(n / coverage_fraction)
        
    lon, lat = [], []
    count = 0
    while len(lon) < n:
        lon_throw = np.random.uniform(0., 360., n_throw)
        lat_throw = np.degrees(np.arcsin(np.random.uniform(-1., 1., n_throw)))

        pix_throw = ugali.utils.healpix.angToPix(nside_pix, lon_throw, lat_throw)
        cut = mask[pix_throw].astype(bool)

        lon = np.append(lon, lon_throw[cut])
        lat = np.append(lat, lat_throw[cut])

        count += 1
        if count > 10:
            raise RuntimeError('Too many loops...')

    return lon[0:n], lat[0:n], area

############################################################

def randomPositionsMask(mask, nside_pix, n):
    """
    Generate n random positions within a HEALPix mask of booleans.

    KCB: Likely superceded by the randomPositions function, but more generic.
    """
    
    npix = len(mask)
    nside = hp.npix2nside(npix)

    # Estimate the number of points that need to be thrown based off
    # coverage fraction of the HEALPix mask
    coverage_fraction = float(np.sum(mask)) / len(mask) 
    n_throw = int(n / coverage_fraction)
        
    lon, lat = [], []
    latch = True
    count = 0
    while len(lon) < n:
        lon_throw = np.random.uniform(0., 360., n_throw)
        lat_throw = np.degrees(np.arcsin(np.random.uniform(-1., 1., n_throw)))

        pix = ugali.utils.healpix.angToPix(nside, lon_throw, lat_throw)
        cut = mask[pix].astype(bool)

        lon = np.append(lon, lon_throw[cut])
        lat = np.append(lat, lat_throw[cut])

        count += 1
        if count > 10:
            raise RuntimeError('Too many loops...')

    return lon[0:n], lat[0:n]

############################################################



#!/usr/bin/env python
"""
Script to automate the process of generating an isochrone library
using the Dotter isochrones.

Download isochrones from:
http://stellar.dartmouth.edu/models/isolf_new.html

New MESA isochrones come from:
http://waps.cfa.harvard.edu/MIST/interp_isos.html
"""

import os
import re
try:
    from urllib.parse import urlencode
    from urllib.request import urlopen
except ImportError:
    from urllib import urlencode
    from urllib2 import urlopen

import requests
import sys
import copy
import tempfile
import subprocess
import shutil
from multiprocessing import Pool
from collections import OrderedDict as odict

import numpy
import numpy as np

import ugali.utils.logger
import ugali.utils.shell
from ugali.utils.logger import logger
from ugali.utils.shell import mkdir
import ugali.analysis.isochrone as iso
from ugali.preprocess.padova import Download

"""
See Vargas et al. 2013 for the distribution of alpha elements in
dSphs: http://adsabs.harvard.edu/abs/2013ApJ...767..134V

Josh Simon remarks: For stars at [Fe/H] > -2, [a/Fe] tends to be
around zero. [Note, though, that this paper does not attempt to do
any membership classification, it just accepts the lists from
Simon & Geha 2007.  I suspect now that we were not sufficiently
conservative on selecting members in those early days, and so some
of the relatively metal-rich stars may in fact be foreground Milky
Way stars.]  More metal-poor stars tend to average more like
[a/Fe] = 0.4-0.5.  Fig. 5 of Frebel et al. (2014) shows similar
plots for individual elements from high-resolution spectra.  Given
these data, plus the empirical fact that the mean metallicities of
the ultra-faint dwarfs are almost universally [Fe/H] < -2, I guess
I would say [a/Fe] = 0.3 is probably the best compromise.

From ADW: Other isochrone sets impose [a/Fe] = 0. For an accurate
comparison between isochrones, I suggest that we stick to [a/Fe] = 0
for the Dotter2008 isochrones as well.

ADW: The Dotter 2008 isochrones are interpolated from a relative
sparse grid of metallicities [-2.5, -2.0, -1.5, -1.0, -0.5, 0.0]. This
can lead to rather large differences between the input and output Z ->
[Fe/H] conversions. We were initially taking the output Zeff value,
but we now use the input Z value for internal consistency.
"""


###########################################################
# Dartmouth Isochrones
# http://stellar.dartmouth.edu/models/isolf_new.php

dict_clr = {'des' : 14,
            'sdss': 11,
            'ps1' : 12,
            }

dict_hel = {'Y=0.245+1.5*Z' : 1,
            'Y=0.33'        : 2,
            'Y=0.40'        : 3}

dict_afe = {'-0.2'            : 1,
            '0 (scaled-solar)': 2,
            '+0.2'            : 3,
            '+0.4'            : 4,
            '+0.6'            : 5,
            '+0.8'            : 6}

dartmouth_defaults = {
    'int':'1', # interpolation: cubic=1, linear=2 (ADW: cubic more accurate)
    'out':'1', # outpur: iso=1, iso+LF=2
    'age':10, # age [Gyr]
    'feh':-2.0, # feh [-2.5 to 0.0]
    'hel': dict_hel['Y=0.245+1.5*Z'], # initial helium abundance
    'afe': dict_afe['0 (scaled-solar)'], # alpha enhancement
    'clr': dict_clr['des'], # photometric system
    'flt':'',
    'bin':'',
    'imf':1,
    'pls':'',
    'lnm':'',
    'lns':'', 
    }

###########################################################
# MESA Isochrones
# http://waps.cfa.harvard.edu/MIST/iso_form.php

# survey system
dict_output = odict([
        ('des' ,'DECam'),
        ('sdss','SDSSugriz'),
        ('ps1' ,'PanSTARRS'),
	('lsst','LSST'),
])

mesa_defaults = {
        'version':'1.0',
        'v_div_vcrit':'vvcrit0.4',
        'age_scale':'linear',
        'age_type':'single',
        'age_value':10e9, # yr if scale='linear'; log10(yr) if scale='log10'
        'age_range_low':'',
        'age_range_high':'',
        'age_range_delta':'',
        'age_list':'',
        'FeH_value':-3.0,
        'theory_output':'basic',
        'output_option':'photometry',
        'output':'DECam',
        'Av_value':0,
}

mesa_defaults_10 = dict(mesa_defaults,version='1.0')

class Dotter2008(Download):
    """ Dartmouth isochrones from Dotter et al. 2008:
    http://stellar.dartmouth.edu/models/isolf_new.html
    """
    defaults = copy.deepcopy(dartmouth_defaults)
    isochrone=iso.Dotter2008

    abins = np.arange(1., 13.5 + 0.1, 0.1)
    zbins = np.arange(7e-5,1e-3 + 1e-5,1e-5)

    def query_server(self, outfile, age, metallicity):
        z = metallicity
        feh = self.isochrone.z2feh(z)
        
        params = dict(self.defaults)
        params['age']=age
        params['feh']='%.6f'%feh
        params['clr']=dict_clr[self.survey]

        url = 'http://stellar.dartmouth.edu/models/isolf_new.php'
        query = url + '?' + urlencode(params)
        logger.debug(query)
        response = urlopen(query)
        page_source = response.read()
        try:
            file_id = int(page_source.split('tmp/tmp')[-1].split('.iso')[0])
        except Exception as e:
            logger.debug(str(e))
            msg = 'Output filename not found'
            raise RuntimeError(msg)

        infile = 'http://stellar.dartmouth.edu/models/tmp/tmp%s.iso'%(file_id)
        command = 'wget -q %s -O %s'%(infile, outfile)
        subprocess.call(command,shell=True)        

        ## ADW: Old code to rename the output file based on Zeff ([a/Fe] corrected)
        #tmpfile = tempfile.NamedTemporaryFile().name
        #tmp = open(tmpfile,'r')
        #lines = [tmp.readline() for i in range(4)]
        #z_eff = float(lines[3].split()[4])
        #basename = self.params2filename(age,z_eff)
     
        #logger.info("Writing %s..."%outfile)
        #mkdir(outdir)
        #shutil.move(tmpfile,outfile)

    @classmethod
    def verify(cls, filename, survey, age, metallicity):
        nlines=8
        with open(filename,'r') as f:
            lines = [f.readline() for i in range(nlines)]

            if len(lines) < nlines:
                msg = "Incorrect file size"
                raise Exception(msg)

            try:
                z = lines[3].split()[4]
                assert np.allclose(metallicity,float(z),atol=1e-3)
            except:
                msg = "Metallicity does not match:\n"+lines[3]
                raise Exception(msg)

            try:
                s = lines[5].split()[2]
                assert dict_output[survey][:4] in s
            except:
                msg = "Incorrect survey:\n"+lines[5]
                raise Exception(msg)

            try:
                a = lines[7].split('=')[1].strip().split()[0]
                assert np.allclose(age,float(a),atol=1e-5)
            except:
                msg = "Age does not match:\n"+lines[7]
                raise Exception(msg)

class Dotter2016(Download):
    """ MESA isochrones from Dotter 2016:
    http://waps.cfa.harvard.edu/MIST/iso_form.php
    """
    defaults = copy.deepcopy(mesa_defaults_10)
    isochrone = iso.Dotter2016

    abins = np.arange(1., 13.5 + 0.1, 0.1)
    zbins = np.arange(1e-5,1e-3 + 1e-5,1e-5)

    def query_server(self, outfile, age, metallicity):
        z = metallicity
        feh = self.isochrone.z2feh(z)
        
        params = dict(self.defaults)
        params['output'] = dict_output[self.survey]
        params['FeH_value'] = feh
        params['age_value'] = age * 1e9
        if params['age_scale'] == 'log10':
            params['age_value'] = np.log10(params['age_value'])

        server = 'http://waps.cfa.harvard.edu/MIST'
        url = server + '/iso_form.php'
        logger.debug("Accessing %s..."%url)
        response = requests.post(url,data=params)

        try:
            fname = os.path.basename(response.text.split('"')[1])
        except Exception as e:
            logger.debug(str(e))
            msg = 'Output filename not found'
            raise RuntimeError(msg)
            
        tmpdir = os.path.dirname(tempfile.NamedTemporaryFile().name)
        tmpfile = os.path.join(tmpdir,fname)

        out = '{0}/tmp/{1}'.format(server, fname)
        cmd = 'wget %s -P %s'%(out,tmpdir)
        logger.debug(cmd)
        stdout = subprocess.check_output(cmd,shell=True,
                                         stderr=subprocess.STDOUT)
        logger.debug(stdout)

        cmd = 'unzip %s -d %s'%(tmpfile,tmpdir)
        logger.debug(cmd)
        stdout = subprocess.check_output(cmd,shell=True,
                                         stderr=subprocess.STDOUT)
        logger.debug(stdout)

        logger.debug("Creating %s..."%outfile)
        shutil.move(tmpfile.replace('.zip','.cmd'),outfile)
        os.remove(tmpfile)

        return outfile

    @classmethod
    def verify(cls, filename, survey, age, metallicity):
        age = age*1e9
        nlines=14
        with open(filename,'r') as f:
            lines = [f.readline() for i in range(nlines)]
            if len(lines) < nlines:
                msg = "Incorrect file size"
                raise Exception(msg)
                
            try:
                s = lines[2].split()[-2]
                assert dict_output[survey][:4] in s
            except:
                msg = "Incorrect survey:\n"+lines[2]
                raise Exception(msg)

            try:
                z = lines[5].split()[2]
                assert np.allclose(metallicity,float(z),atol=1e-3)
            except:
                msg = "Metallicity does not match:\n"+lines[5]
                raise Exception(msg)

            try:
                a = lines[13].split()[1]
                assert np.allclose(age,float(a),atol=1e-5)
            except:
                msg = "Age does not match:\n"+lines[13]
                raise Exception(msg)

def factory(name, **kwargs):
    from ugali.utils.factory import factory
    return factory(name, module=__name__, **kwargs)

if __name__ == "__main__":
    import ugali.utils.parser
    description = "Download Dotter isochrones"
    parser = ugali.utils.parser.Parser(description=description)
    parser.add_verbose()
    parser.add_force()
    parser.add_argument('-a','--age',default=None,type=float,action='append')
    parser.add_argument('-z','--metallicity',default=None,type=float,action='append')
    parser.add_argument('-k','--kind',default='Dotter2008')
    parser.add_argument('-s','--survey',default='des')
    parser.add_argument('-o','--outdir',default=None)
    parser.add_argument('-n','--njobs',default=1,type=int)
    args = parser.parse_args()

    if args.verbose:
        try:
            from http.client import HTTPConnection
        except ImportError:
            from httplib import HTTPConnection
        HTTPConnection.debuglevel = 1

    if args.outdir is None: 
        args.outdir = os.path.join(args.survey.lower(),args.kind.lower())
    logger.info("Writing to output directory: %s"%args.outdir)

    p = factory(args.kind,survey=args.survey)

    abins = args.age if args.age else p.abins
    zbins = args.metallicity if args.metallicity else p.zbins
    grid = [g.flatten() for g in np.meshgrid(abins,zbins)]
    logger.info("Ages:\n  %s"%np.unique(grid[0]))
    logger.info("Metallicities:\n  %s"%np.unique(grid[1]))

    def run(args):
        try:  
            p.download(*args)
        except Exception as e:
            # power through any exceptions...
            logger.warn(str(e))

    arguments = [(a,z,args.outdir,args.force) for a,z in zip(*grid)]
    if args.njobs > 1:
        msg = "Multiprocessing does not work for %s download."%args.kind
        raise Exception(msg)
    #elif args.njobs > 1:
    #    pool = Pool(processes=args.njobs, maxtasksperchild=100)
    #    results = pool.map(run,arguments)
    else:
        results = list(map(run,arguments))



#!/usr/bin/env python
"""
Cross-platform batch computing interface.
"""
import os
import subprocess, subprocess as sub
import getpass
from collections import OrderedDict as odict
from itertools import chain
import copy
import time
import resource

import numpy as np

from ugali.utils.logger import logger

GB=1024**3
CLUSTERS = odict([
    ('local' ,['local']),
    ('lsf'   ,['lsf','slac','kipac']),
    ('slurm' ,['slurm','midway','kicp']),
    ('condor',['condor','fnal']),
])

QUEUES = odict([
    ('local' ,[]),
    ('lsf'   ,['express','short','medium','long','xlong','xxl',
               'kipac-ibq','bulletmpi']),
    ('slurm' ,[]),
    ('condor',['local','vanilla','universe','grid']),
])

# https://confluence.slac.stanford.edu/x/OaUlCw
RUNLIMITS = odict([              #Hard limits
        (None       ,'4:00'),    # Default value
        ('express'  ,'0:04'),    # 0:04
        ('short'    ,'0:59'),    # 1:00
        ('medium'   ,'1:00'),    # 48:00
        ('long'     ,'4:00'),    # 120:00
        ('xlong'    ,'72:00'),   # 72:00
        ('xxl'      ,'168:00'),
        # MPI queues
        ('kipac-ibq','36:00'),   # 24:00 (deprecated)
        ('bulletmpi','36:00'),   # 72:00
        ])

# SLAC updated how memory was handled on the batch system  
# General queues now only have 4GB of RAM per CPU
# To get more memory for a general job, you need to request more cores:
# i.e., '-n 2 -R "span[hosts=1]"' for 8GB of RAM
# https://confluence.slac.stanford.edu/x/moNdCw

# These are options for MPI jobs 
# '-R "span[ptile=16]"' indicates the number of processors on each
# host that should be allocated to the job
MPIOPTS = odict([
        (None       ,' -R "span[ptile=4]"'),
        ('local'    ,''),
        ('short'    ,''),
        ('medium'   ,''),
        ('long'     ,''),
        ('kipac-ibq',' -R "span[ptile=8]"'),
        #('bulletmpi',' -R "span[ptile=16]"'),
        ('bulletmpi',''),
        ])

def factory(queue,**kwargs):
    # The default for the factory
    if queue is None: queue = 'local'

    name = queue.lower()
    if name in list(chain(*list(QUEUES.values()))):
        kwargs.setdefault('q',name)

    if name in CLUSTERS['local']+QUEUES['local']:
        batch = Local(**kwargs)
    elif name in CLUSTERS['lsf']+QUEUES['lsf']:
        batch = LSF(**kwargs)
    elif name in CLUSTERS['slurm']+QUEUES['slurm']:
        batch = Slurm(**kwargs)
    elif name in CLUSTERS['condor']+QUEUES['condor']:
        # Need to learn how to use condor first...
        batch = Condor(**kwargs)
        raise Exception("Condor cluster not implemented")
    else:
        raise TypeError('Unexpected queue name: %s'%name)

    return batch

batchFactory = factory
batch_factory = factory

class Batch(object):
    # Default options for batch submission
    _defaults = odict([])
    # Map between generic and batch specific names
    _mapping = odict([])

    def __init__(self, **kwargs):
        self.username = getpass.getuser()
        self.max_jobs = kwargs.pop('max_jobs',None)
        self.default_opts = copy.deepcopy(self._defaults)
        self.default_opts.update(**kwargs)
        self.submit_cmd = "submit %(opts)s %(command)s"
        self.jobs_cmd = "jobs"

    def jobs(self):
        out = self.popen(self.jobs_cmd)
        stdout = out.communicate()[0]
        return stdout

    def njobs(self):
        # Remove header line
        jobs = self.jobs()
        return len(jobs.strip().split('\n'))-1 if jobs else 0

    def bfail(self, path):
        """ Check logfile(s) for failure. 

        Parameters:
        -----------
        path : path to logfile(s)
        
        Returns:
        --------
        files : failed logfiles and message
        """
        cmd='grep -r "^Exited" %s'%path
        out = self.popen(cmd)
        stdout = out.communicate()[0]
        # There is an extra newline at the end
        return stdout.split('\n')[:-1]

    def bcomp(self, path):
        """ Return completed logfile(s).

        Parameters:
        -----------
        path : path to logfile(s)
        
        Returns:
        --------
        files : completed logfiles and message
        """
        cmd='grep -r "^Successfully completed." %s'%path
        out = self.popen(cmd)
        stdout = out.communicate()[0]
        # There is an extra newline at the end
        return stdout.split('\n')[:-1]

    def throttle(self,max_jobs=None,sleep=60):
        if max_jobs is None: max_jobs = self.max_jobs
        if max_jobs is None: return
        while True:
            njobs = self.njobs()
            if njobs < max_jobs:
                return
            else:
                logger.info('%i jobs already in queue, waiting...'%(njobs))
                time.sleep(sleep)

    def popen(self, command):
        return sub.Popen(command,shell=True,
                         stdin=sub.PIPE,stdout=sub.PIPE,stderr=sub.PIPE)

    def call(self, command):
        logger.debug(command)
        return sub.call(command,shell=True)

    def remap_options(self,opts):
        for k in self._mapping.keys():
            v = opts.pop(k,None)
            if v is not None:
                opts[self._mapping[k]] = v

    def batch(self, command, jobname=None, logfile=None, **opts):
        if jobname: opts.update(jobname=jobname)
        if logfile: opts.update(logfile=logfile)
        self.remap_options(opts)
        params = dict(opts=self.parse_options(**opts),command=command)
        return self.submit_cmd%params

    def submit(self, command, jobname=None, logfile=None, **opts):
        cmd = self.batch(command, jobname, logfile, **opts)
        self.throttle()
        self.call(cmd)
        return cmd

    def get_memory_limit(): pass
    def set_memory_limit(mlimit): pass

class Local(Batch):
    def __init__(self,**kwargs):
        super(Local,self).__init__(**kwargs)
        self.jobs_cmd = "echo 0"
        self.submit_cmd = "%(command)s %(opts)s"

    def parse_options(self,**opts):
        if opts.get('logfile'): return ' 2>&1 | tee %(logfile)s'%opts
        return ''

    def njobs(self):
        return 0

class LSF(Batch):
    _defaults = odict([
        ('R','"scratch > 1 && rhel60"'),
        ('C', 0),
        #('M','8G'),
        #('q', 'long'),
    ])

    _mapping = odict([
        ('jobname','J'),
        ('logfile','oo')
    ])

    def __init__(self,**kwargs):
        super(LSF,self).__init__(**kwargs)

        self.jobs_cmd = "bjobs -u %s"%self.username
        self.submit_cmd = "bsub %(opts)s %(command)s"

    def runlimit(self, queue=None):
        """
        Translate queue to wallclock runlimit.
        """
        try:             return RUNLIMITS[queue]
        except KeyError: return RUNLIMITS[None]
    q2w = runlimit

    def mpiopts(self, queue=None):
        """
        Translate queue into MPI options.
        """
        try:             return MPIOPTS[queue]
        except KeyError: return MPIOPTS[None]
    q2mpi = mpiopts

    def parse_options(self, **opts):
        # Default options for the cluster
        options = odict(self.default_opts)
        # Default options for the queue
        #options.update(OPTIONS[options.get('q')])
        # User specified options
        options.update(opts)
        #if 'n' in list(options.keys()): 
        #    #options['a'] = 'mpirun'
        #    options['R'] += self.mpiopts(options.get('q'))
        options.setdefault('W',self.runlimit(options.get('q')))
        return ''.join('-%s %s '%(k,v) for k,v in options.items())

    @classmethod
    def set_memory_limit(cls, mlimit):
        """Set the (soft) memory limit for setrlimit.
        
        Parameters:
        -----------
        mlimit : soft memory limit (bytes)
        
        Returns:
        --------
        soft, hard : memory limits (bytes)
        """
        rsrc = resource.RLIMIT_AS
        resource.setrlimit(rsrc, (mlimit, mlimit))
        return resource.getrlimit(rsrc)

    @classmethod
    def get_memory_limit(cls):
        """Get the hard memory limit from LSF.

        Parameters
        ----------
        None
        
        Returns
        -------
        mlimit : memory limit (bytes)
        """
        rsrc = resource.RLIMIT_AS
        soft, hard = resource.getrlimit(rsrc)
        if os.getenv('LSB_CG_MEMLIMIT') and os.getenv('LSB_HOSTS'):
            # Memory limit per core
            memlimit = int(os.getenv('LSB_CG_MEMLIMIT'), 16)
            # Number of cores
            ncores = len(os.getenv('LSB_HOSTS').split())
            #soft = ncores * memlimit - 100*1024**2
            soft = int( 0.95 * ncores * memlimit )
        return soft,hard

        
class Slurm(Batch):
    _defaults = odict([
        ('account','kicp'),
        ('partition','kicp-ht'),
        ('mem',10000)
    ])

    def __init__(self, **kwargs):
        super(Slurm,self).__init__(**kwargs)
        logger.warning('Slurm cluster is untested')

        self.jobs_cmd = "squeue -u %s"%self.username
        self.submit_cmd = "sbatch %(opts)s %(command)s"

    def parse_options(self, **opts):
        options = odict(self.default_opts)
        options.update(opts)
        return ''.join('--%s %s '%(k,v) for k,v in options.items())

class Condor(Batch):
    """ Not implemented yet... """
    def __init__(self, **kwargs):
        super(Condor,self).__init__(**kwargs)
        logger.warning('Condor cluster is untested')
        
        self.jobs_cmd = 'condor_q -u %s'%self.username
        self.submit_cmd = "csub %(opts)s %(command)s"

if __name__ == "__main__":
    import argparse
    description = __doc__
    parser = argparse.ArgumentParser(description=description)
    args = parser.parse_args()




#!/usr/bin/env python
"""
Toolkit for working with healpix
"""

from collections import OrderedDict as odict

import numpy
import numpy as np
import healpy as hp
import fitsio

import ugali.utils.projector
from ugali.utils import fileio
from ugali.utils.logger import logger
import ugali.utils.fileio

############################################################

def superpixel(subpix, nside_subpix, nside_superpix):
    """
    Return the indices of the super-pixels which contain each of the sub-pixels.
    """
    if nside_subpix==nside_superpix: return subpix
    theta, phi =  hp.pix2ang(nside_subpix, subpix)
    return hp.ang2pix(nside_superpix, theta, phi)


def subpixel(superpix, nside_superpix, nside_subpix):
    """
    Return the indices of sub-pixels (resolution nside_subpix) within
    the super-pixel with (resolution nside_superpix).
    
    ADW: It would be better to convert to next and do this explicitly
    """
    if nside_superpix==nside_subpix: return superpix
    vec = hp.pix2vec(nside_superpix, superpix)
    radius = np.degrees(2. * hp.max_pixrad(nside_superpix))
    subpix = query_disc(nside_subpix, vec, radius)
    pix_for_subpix = superpixel(subpix,nside_subpix,nside_superpix)
    # Might be able to speed up array indexing...
    return subpix[pix_for_subpix == superpix]

def d_grade_ipix(ipix, nside_in, nside_out, nest=False):
    """
    Return the indices of the super-pixels which contain each of the
    sub-pixels (nside_in > nside_out).

    Parameters:
    -----------
    ipix      : index of the input subpixels
    nside_in  : nside of the input subpix
    nside_out : nside of the desired superpixels

    Returns:
    --------
    ipix_out  : superpixels for each subpixel
    """

    if nside_in==nside_out: return ipix
    if not (nside_in > nside_out): 
        raise ValueError("nside_out must be less than nside_in")

    return hp.vec2pix(nside_out, *hp.pix2vec(nside_in, ipix, nest), nest=nest)

def u_grade_ipix(ipix, nside_in, nside_out, nest=False):
    """
    Return the indices of sub-pixels (resolution nside_subpix) within
    the super-pixel(s) (resolution nside_superpix).
    
    Parameters:
    -----------
    ipix      : index of the input superpixel(s)
    nside_in  : nside of the input superpixel
    nside_out : nside of the desired subpixels

    Returns:
    --------
    ipix_out : subpixels for each superpixel
    """

    if nside_in==nside_out: return ipix
    if not (nside_in < nside_out): 
        raise ValueError("nside_in must be less than nside_out")

    if nest: nest_ipix = ipix
    else:    nest_ipix = hp.ring2nest(nside_in, ipix)

    factor = (nside_out//nside_in)**2
    if np.isscalar(ipix):
        nest_ipix_out = factor*nest_ipix + np.arange(factor)
    else:
        nest_ipix_out = factor*np.asarray(nest_ipix)[:,np.newaxis]+np.arange(factor)

    if nest: return nest_ipix_out
    else:    return hp.nest2ring(nside_out, nest_ipix_out)
        

def ud_grade_ipix(ipix, nside_in, nside_out, nest=False):
    """
    Upgrade or degrade resolution of a pixel list.

    Parameters:
    -----------
    ipix:array-like 
    the input pixel(s)

    nside_in:int
    the nside of the input pixel(s)

    nside_out:int
    the desired nside of the output pixel(s)

    order:str
    pixel ordering of input and output ("RING" or "NESTED")

    Returns:
    --------
    pix_out:array-like
    the upgraded or degraded pixel array
    """
    if nside_in == nside_out: return ipix
    elif nside_in < nside_out:
        return u_grade_ipix(ipix, nside_in, nside_out, nest)
    elif nside_in > nside_out:
        return d_grade_ipix(ipix, nside_in, nside_out, nest)

############################################################

#ADW: These can be replaced by healpy functions...

def phi2lon(phi): return np.degrees(phi)
def lon2phi(lon): return np.radians(lon)

def theta2lat(theta): return 90. - np.degrees(theta)
def lat2theta(lat): return np.radians(90. - lat)

def pix2ang(nside, pix, nest=False):
    """
    Return (lon, lat) in degrees instead of (theta, phi) in radians
    """
    theta, phi =  hp.pix2ang(nside, pix, nest=nest)
    lon = phi2lon(phi)
    lat = theta2lat(theta)
    return lon, lat

def ang2pix(nside, lon, lat, nest=False):
    """
    Input (lon, lat) in degrees instead of (theta, phi) in radians
    """
    theta = np.radians(90. - lat)
    phi = np.radians(lon)
    return hp.ang2pix(nside, theta, phi, nest=nest)

def ang2vec(lon, lat):
    theta = lat2theta(lat)
    phi = lon2phi(lon)
    vec = hp.ang2vec(theta, phi)
    return vec

pixToAng = pix2ang
angToPix = ang2pix
angToVec = ang2vec

def get_nside(m):
    try: 
        return hp.get_nside(m)
    except TypeError:
        return hp.get_nside(m.data)

############################################################

def healpixMap(nside, lon, lat, fill_value=0., nest=False):
    """
    Input (lon, lat) in degrees instead of (theta, phi) in radians.
    Returns HEALPix map at the desired resolution 
    """

    lon_median, lat_median = np.median(lon), np.median(lat)
    max_angsep = np.max(ugali.utils.projector.angsep(lon, lat, lon_median, lat_median))
    
    pix = angToPix(nside, lon, lat, nest=nest)
    if max_angsep < 10:
        # More efficient histograming for small regions of sky
        m = np.tile(fill_value, hp.nside2npix(nside))
        pix_subset = ang2disc(nside, lon_median, lat_median, max_angsep, nest=nest)
        bins = np.arange(np.min(pix_subset), np.max(pix_subset) + 1)
        m_subset = np.histogram(pix, bins=bins - 0.5)[0].astype(float)
        m[bins[0:-1]] = m_subset
    else:
        m = np.histogram(pix, np.arange(hp.nside2npix(nside) + 1))[0].astype(float)
    if fill_value != 0.:
        m[m == 0.] = fill_value
    return m

############################################################

def in_pixels(lon,lat,pixels,nside):
    """
    Check if (lon,lat) in pixel list. Assumes RING formatting.

    Parameters:
    -----------
    lon    : longitude (deg)
    lat    : latitude (deg)
    pixels : pixel list [RING format] to check for inclusion
    nside  : nside of pixel list 

    Returns:
    --------
    inpix : boolean array for inclusion
    """
    pix = ang2pix(nside,lon,lat)
    return np.in1d(pix,pixels)


def index_pix_in_pixels(pix,pixels,sort=False,outside=-1):
    """
    Find the indices of a set of pixels into another set of pixels.
    !!! ASSUMES SORTED PIXELS !!!

    Parameters:
    -----------
    pix    : set of search pixels
    pixels : set of reference pixels
  
    Returns:
    --------
    index : index into the reference pixels
    """
    # ADW: Not really safe to set index = -1 (accesses last entry); 
    # -np.inf would be better, but breaks other code...

    # ADW: Are the pixels always sorted? Is there a quick way to check?
    if sort: pixels = np.sort(pixels)

    # Assumes that 'pixels' is pre-sorted, otherwise...???
    index = np.searchsorted(pixels,pix)
    if np.isscalar(index):
        if not np.in1d(pix,pixels).any(): index = outside
    else:
        # Find objects that are outside the pixels
        index[~np.in1d(pix,pixels)] = outside
    return index

def index_lonlat_in_pixels(lon,lat,pixels,nside,sort=False,outside=-1):
    """
    Find the indices of a set of angles into a set of pixels

    Parameters:
    -----------
    pix    : set of search pixels
    pixels : set of reference pixels
  
    Returns:
    --------
    index : index into the reference pixels
    """

    pix = ang2pix(nside,lon,lat)
    return index_pix_in_pixels(pix,pixels,sort,outside)

index_pixels = index_lonlat_in_pixels

#def index_pixels(lon,lat,pixels,nside):
#    """
#    Find the index for object amoung a subset of healpix pixels.
#    Set index of objects outside the pixel subset to -1
#    """
#    # ADW: Not really safe to set index = -1 (accesses last entry); 
#    # -np.inf would be better, but breaks other code...
#    pix = ang2pix(nside,lon,lat)
#    # pixels should be pre-sorted, otherwise...???
#    index = np.searchsorted(pixels,pix)
#    if np.isscalar(index):
#        if not np.in1d(pix,pixels).any(): index = -1
#    else:
#        # Find objects that are outside the roi
#        #index[np.take(pixels,index,mode='clip')!=pix] = -1
#        index[~np.in1d(pix,pixels)] = -1
#    return index


############################################################

def query_disc(nside, vec, radius, inclusive=False, fact=4, nest=False):
    """
    Wrapper around healpy.query_disc to deal with old healpy implementation.

    nside : int
      The nside of the Healpix map.
    vec : float, sequence of 3 elements
      The coordinates of unit vector defining the disk center.
    radius : float
      The radius (in degrees) of the disc
    inclusive : bool, optional
      If False, return the exact set of pixels whose pixel centers lie 
      within the disk; if True, return all pixels that overlap with the disk,
      and maybe a few more. Default: False
    fact : int, optional
      Only used when inclusive=True. The overlapping test will be done at
      the resolution fact*nside. For NESTED ordering, fact must be a power of 2,
      else it can be any positive integer. Default: 4.
    nest: bool, optional
      if True, assume NESTED pixel ordering, otherwise, RING pixel ordering

    """
    try: 
        # New-style call (healpy 1.6.3)
        return hp.query_disc(nside, vec, np.radians(radius), inclusive, fact, nest)
    except Exception as e: 
        print(e)
        # Old-style call (healpy 0.10.2)
        return hp.query_disc(nside, vec, np.radians(radius), nest, deg=False)

def ang2disc(nside, lon, lat, radius, inclusive=False, fact=4, nest=False):
    """
    Wrap `query_disc` to use lon, lat, and radius in degrees.
    """
    vec = ang2vec(lon,lat)
    return query_disc(nside,vec,radius,inclusive,fact,nest)

angToDisc = ang2disc

def get_interp_val(m, lon, lat, *args, **kwargs):
    return hp.get_interp_val(m, lat2theta(lat), lon2phi(lon), *args, **kwargs)

############################################################

def header_odict(nside,nest=False,coord=None, partial=True):
    """Mimic the healpy header keywords."""
    hdr = odict([])
    hdr['PIXTYPE']=odict([('name','PIXTYPE'),
                          ('value','HEALPIX'),
                          ('comment','HEALPIX pixelisation')])

    ordering = 'NEST' if nest else 'RING'
    hdr['ORDERING']=odict([('name','ORDERING'),
                           ('value',ordering),
                           ('comment','Pixel ordering scheme, either RING or NESTED')])
    hdr['NSIDE']=odict([('name','NSIDE'),
                        ('value',nside),
                        ('comment','Resolution parameter of HEALPIX')])
    if coord:
        hdr['COORDSYS']=odict([('name','COORDSYS'), 
                               ('value',coord), 
                               ('comment','Ecliptic, Galactic or Celestial (equatorial)')])
    
    if not partial:
        hdr['FIRSTPIX']=odict([('name','FIRSTPIX'),
                               ('value',0), 
                               ('comment','First pixel # (0 based)')])
        hdr['LASTPIX']=odict([('name','LASTPIX'),
                              ('value',hp.nside2npix(nside)-1),
                              ('comment','Last pixel # (0 based)')])
    hdr['INDXSCHM']=odict([('name','INDXSCHM'),
                           ('value','EXPLICIT' if partial else 'IMPLICIT'),
                           ('comment','Indexing: IMPLICIT or EXPLICIT')])
    hdr['OBJECT']=odict([('name','OBJECT'), 
                         ('value','PARTIAL' if partial else 'FULLSKY'),
                         ('comment','Sky coverage, either FULLSKY or PARTIAL')])
    return hdr

def write_partial_map(filename, data, nside, coord=None, nest=False,
                      header=None,dtype=None,**kwargs):
    """
    Partial HEALPix maps are used to efficiently store maps of the sky by only
    writing out the pixels that contain data.

    Three-dimensional data can be saved by supplying a distance modulus array
    which is stored in a separate extension.

    Parameters:
    -----------
    filename : output file name
    data     : dictionary or recarray of data to write (must contain 'PIXEL')
    nside    : healpix nside of data
    coord    : 'G'alactic, 'C'elestial, 'E'cliptic
    ordering : 'RING' or 'NEST'
    kwargs   : Passed to fitsio.write

    Returns:
    --------
    None
    """

    # ADW: Do we want to make everything uppercase?

    if isinstance(data,dict):
        names = list(data.keys())
    else:
        names = data.dtype.names

    if 'PIXEL' not in names:
        msg = "'PIXEL' column not found."
        raise ValueError(msg)

    hdr = header_odict(nside=nside,coord=coord,nest=nest)
    fitshdr = fitsio.FITSHDR(list(hdr.values()))
    if header is not None:
        for k,v in header.items():
            fitshdr.add_record({'name':k,'value':v})
    # ADW: Should this be a debug?
    logger.info("Writing %s..."%filename)
    fitsio.write(filename,data,extname='PIX_DATA',header=fitshdr,clobber=True)

def read_partial_map(filenames, column, fullsky=True, **kwargs):
    """
    Read a partial HEALPix file(s) and return pixels and values/map. Can
    handle 3D healpix maps (pix, value, zdim). Returned array has
    shape (dimz,npix).

    Parameters:
    -----------
    filenames     : list of input filenames
    column        : column of interest
    fullsky       : partial or fullsky map
    kwargs        : passed to fitsio.read

    Returns:
    --------
    (nside,pix,map) : pixel array and healpix map (partial or fullsky)
    """
    import fitsio

    # Make sure that PIXEL is in columns
    #kwargs['columns'] = ['PIXEL',column]
    kwargs['columns'] = ['PIXEL'] + np.atleast_1d(column).tolist()

    filenames = np.atleast_1d(filenames)
    header = fitsio.read_header(filenames[0],ext=kwargs.get('ext',1))
    data = fileio.load_files(filenames,**kwargs)

    pix = data['PIXEL']
    value = data[column]
    nside = header['NSIDE']
    npix = hp.nside2npix(nside)

    ndupes = len(pix) - len(np.unique(pix))
    if ndupes > 0:
        msg = '%i duplicate pixels during load.'%(ndupes)
        raise Exception(msg)

    if fullsky and not np.isscalar(column):
        raise Exception("Cannot make fullsky map from list of columns.")
    
    if fullsky:
        shape = list(value.shape)
        shape[0] = npix
        hpxmap = hp.UNSEEN * np.ones(shape,dtype=value.dtype)
        hpxmap[pix] = value
        return (nside,pix,hpxmap.T)
    else:
        return (nside,pix,value.T)

def merge_partial_maps(filenames,outfile,**kwargs):
    """
    Parameters:
    -----------
    filenames : list of filenames to merge
    outfile   : name of outfile to create
    kwargs    : kwargs passed to fitsio.read

    Returns:
    --------
    None
    """
    import fitsio
    filenames = np.atleast_1d(filenames)

    header = fitsio.read_header(filenames[0],ext=kwargs.get('ext',1))
    nside = header['NSIDE']
    data = fileio.load_files(filenames,**kwargs)
    pix = data['PIXEL']

    ndupes = len(pix) - len(np.unique(pix))
    if ndupes > 0:
        msg = '%i duplicate pixels during load.'%(ndupes)
        raise Exception(msg)

    extname = 'DISTANCE_MODULUS'
    kwargs['ext'] = extname
    distance = fileio.load_files(filenames,**kwargs)[extname]
    unique_distance = np.unique(distance)
    # Check if distance moduli are the same...
    if np.any(distance[:len(unique_distance)] != unique_distance):
        msg = "Non-matching distance modulus:"
        msg += '\n'+str(distance[:len(unique_distance)])
        msg += '\n'+str(unique_distance)
        raise Exception(msg)
    del distance

    # Writing partial maps
    if outfile is not None:
        write_partial_map(outfile,data=data,nside=nside,clobber=True)
        fitsio.write(outfile,{extname:unique_distance},extname=extname)

    return nside,data,unique_distance

def merge_likelihood_headers(filenames, outfile, **kwargs):
    """
    Merge header information from likelihood files.

    Parameters:
    -----------
    filenames : input filenames
    oufile    : the merged file to write
    
    Returns:
    --------
    data      : the data being written
    """
    import fitsio
    filenames = np.atleast_1d(filenames)

    ext='PIX_DATA'
    nside = fitsio.read_header(filenames[0],ext=ext)['LKDNSIDE']

    keys = ['LKDPIX','STELLAR','NINSIDE','NANNULUS']
    data_dict = fileio.load_headers(filenames,ext=ext,keys=keys,mulitproc=8)
    names = data_dict.dtype.names
    data_dict.dtype.names = ['PIXEL' if n=='LKDPIX' else n for n in names]
        
    write_partial_map(outfile, data_dict, nside)
    return data_dict

def merge_likelihood_headers2(filenames, outfile, **kwargs):
    """
    Merge header information from likelihood files.

    Parameters:
    -----------
    filenames : input filenames
    oufile    : the merged file to write
    
    Returns:
    --------
    data      : the data being written
    """
    import fitsio
    filenames = np.atleast_1d(filenames)

    ext='PIX_DATA'
    nside = fitsio.read_header(filenames[0],ext=ext)['LKDNSIDE']
    
    keys=['STELLAR','NINSIDE','NANNULUS']
    data_dict = odict(PIXEL=[])
    for k in keys:
        data_dict[k] = []

    for i,filename in enumerate(filenames):
        logger.debug('(%i/%i) %s'%(i+1, len(filenames), filename))
        header = fitsio.read_header(filename,ext=ext)
        data_dict['PIXEL'].append(header['LKDPIX'])
        for key in keys:
            data_dict[key].append(header[key])

        del header
        
    data_dict['PIXEL'] = np.array(data_dict['PIXEL'],dtype=int)
    for key in keys:
        data_dict[key] = np.array(data_dict[key],dtype='f4')

    write_partial_map(outfile, data_dict, nside)
    return data_dict


def read_map(filename, nest=False, hdu=None, h=False, verbose=True):
    """Read a healpix map from a fits file.  Partial-sky files,
    if properly identified, are expanded to full size and filled with UNSEEN.
    Uses fitsio to mirror much (but not all) of the functionality of healpy.read_map
    
    Parameters:
    -----------
    filename : str 
      the fits file name
    nest : bool, optional
      If True return the map in NEST ordering, otherwise in RING ordering;
      use fits keyword ORDERING to decide whether conversion is needed or not
      If None, no conversion is performed.
    hdu : int, optional
      the header number to look at (start at 0)
    h : bool, optional
      If True, return also the header. Default: False.
    verbose : bool, optional
      If True, print a number of diagnostic messages
    
    Returns
    -------
    m [, header] : array, optionally with header appended
      The map read from the file, and the header if *h* is True.
    """
    import fitsio
    data,hdr = fitsio.read(filename,header=True,ext=hdu)

    nside = int(hdr.get('NSIDE'))
    if verbose: print('NSIDE = {0:d}'.format(nside))

    if not hp.isnsideok(nside):
        raise ValueError('Wrong nside parameter.')
    sz=hp.nside2npix(nside)

    ordering = hdr.get('ORDERING','UNDEF').strip()
    if verbose: print('ORDERING = {0:s} in fits file'.format(ordering))

    schm = hdr.get('INDXSCHM', 'UNDEF').strip()
    if verbose: print('INDXSCHM = {0:s}'.format(schm))
    if schm == 'EXPLICIT':
        partial = True
    elif schm == 'IMPLICIT':
        partial = False

    # monkey patch on a field method
    fields = data.dtype.names

    # Could be done more efficiently (but complicated) by reordering first
    if hdr['INDXSCHM'] == 'EXPLICIT':
        m = hp.UNSEEN*np.ones(sz,dtype=data[fields[1]].dtype)
        m[data[fields[0]]] = data[fields[1]]
    else:
        m = data[fields[0]].ravel()

    if (not hp.isnpixok(m.size) or (sz>0 and sz != m.size)) and verbose:
        print('nside={0:d}, sz={1:d}, m.size={2:d}'.format(nside,sz,m.size))
        raise ValueError('Wrong nside parameter.')
    if nest is not None:
        if nest and ordering.startswith('RING'):
            idx = hp.nest2ring(nside,np.arange(m.size,dtype=np.int32))
            m = m[idx]
            if verbose: print('Ordering converted to NEST')
        elif (not nest) and ordering.startswith('NESTED'):
            idx = hp.ring2nest(nside,np.arange(m.size,dtype=np.int32))
            m = m[idx]
            if verbose: print('Ordering converted to RING')

    if h: return m, hdr
    else: return m

if __name__ == "__main__":
    import argparse
    description = __doc__
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('args',nargs=argparse.REMAINDER)
    opts = parser.parse_args(); args = opts.args



#!/usr/bin/env python
"""
Older (pre-PARSEC) Padova isochrones.

Not verified to work...
"""
import os.path
from collections import OrderedDict as odict

import numpy as np

from ugali.analysis.model import Model, Parameter
from ugali.utils.shell import mkdir, get_ugali_dir, get_iso_dir
from ugali.isochrone.model import Isochrone
from ugali.isochrone.parsec import Marigo2017 as Padova
from ugali.isochrone.parsec import defaults_27
from ugali.utils.logger import logger

class Girardi2002(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi2000'

class Marigo2008(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'ma08'

class Girardi2010a(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi10a'

class Girardi2010b(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi10b'

class Girardi2002(Padova):
    #_dirname = '/u/ki/kadrlica/des/isochrones/v5/'
    _dirname =  os.path.join(get_iso_dir(),'{survey}','girardi2002')
    # For use with Marigo et al. (2008) and earlier use Anders & Grevesse 1989
    _zsolar = 0.019

    defaults = (Isochrone.defaults) + (
        ('dirname',_dirname,'Directory name for isochrone files'),
        ('hb_stage','BHeb','Horizontal branch stage name'),
        ('hb_spread',0.1,'Intrinisic spread added to horizontal branch'),
        )

    columns = dict(
        des = odict([
                (2, ('mass_init',float)),
                (3, ('mass_act',float)),
                (4, ('log_lum',float)),
                (9, ('g',float)),
                (10, ('r',float)),
                (11,('i',float)),
                (12,('z',float)),
                (13,('Y',float)),
                (15,('stage',object))
                ]),
        )
    
    def _parse(self,filename):
        """
        Reads an isochrone in the old Padova format (Girardi 2002,
        Marigo 2008) and determines the age (log10 yrs and Gyr),
        metallicity (Z and [Fe/H]), and creates arrays with the
        initial stellar mass and corresponding magnitudes for each
        step along the isochrone.
        http://stev.oapd.inaf.it/cgi-bin/cmd
        """
        try:
            columns = self.columns[self.survey.lower()]
        except KeyError as e:
            logger.warning('did not recognize survey %s'%(self.survey))
            raise(e)

        kwargs = dict(delimiter='\t',usecols=list(columns.keys()),dtype=list(columns.values()))
        self.data = np.genfromtxt(filename,**kwargs)
        
        self.mass_init = self.data['mass_init']
        self.mass_act  = self.data['mass_act']
        self.luminosity = 10**self.data['log_lum']
        self.mag_1 = self.data[self.band_1]
        self.mag_2 = self.data[self.band_2]
        self.stage = np.char.array(self.data['stage']).strip()
        for i,s in enumerate(self.stage):
            if i>0 and s=='' and self.stage[i-1]!='':
                self.stage[i] = self.stage[i-1]

        # Check where post-AGB isochrone data points begin
        self.mass_init_upper_bound = np.max(self.mass_init)
        if np.any(self.stage == 'LTP'):
            self.index = np.nonzero(self.stage == 'LTP')[0][0]
        else:
            self.index = len(self.mass_init)

        self.mag = self.mag_1 if self.band_1_detection else self.mag_2
        self.color = self.mag_1 - self.mag_2

class Girardi2010(Girardi2002):
    #_dirname = '/u/ki/kadrlica/des/isochrones/v4/'
    _dirname =  os.path.join(get_iso_dir(),'{survey}','girardi2010')
    _zsolar = 0.019

    defaults = (Isochrone.defaults) + (
        ('dirname',_dirname,'Directory name for isochrone files'),
        ('hb_stage','BHeb','Horizontal branch stage name'),
        ('hb_spread',0.1,'Intrinisic spread added to horizontal branch'),
        )

    columns = dict(
        des = odict([
                (2, ('mass_init',float)),
                (3, ('mass_act',float)),
                (4, ('log_lum',float)),
                (9, ('g',float)),
                (10,('r',float)),
                (11,('i',float)),
                (12,('z',float)),
                (13,('Y',float)),
                (19,('stage',object))
                ]),
        )



"""
Classes to handle initial mass functions (IMFs).

https://github.com/keflavich/imf
"""
from abc import abstractmethod
import numpy as np
import scipy.interpolate

from ugali.utils.logger import logger

############################################################

class IMF(object):
    """
    Base class for initial mass functions (IMFs).
    """

    def __call__(self, mass, **kwargs):
        """ Call the pdf of the mass function """
        return self.pdf(mass,**kwargs)

    def integrate(self, mass_min, mass_max, log_mode=True, weight=False, steps=10000):
        """ Numerical Riemannn integral of the IMF (stupid simple).

        Parameters:
        -----------
        mass_min: minimum mass bound for integration (solar masses)
        mass_max: maximum mass bound for integration (solar masses)
        log_mode[True]: use logarithmic steps in stellar mass as oppose to linear
        weight[False]: weight the integral by stellar mass
        steps: number of numerical integration steps

        Returns:
        --------
        result of integral
        """
        steps = int(steps)
        if log_mode:
            d_log_mass = (np.log10(mass_max) - np.log10(mass_min)) / float(steps)
            log_mass = np.linspace(np.log10(mass_min), np.log10(mass_max), steps)
            mass = 10.**log_mass

            if weight:
                return np.sum(mass * d_log_mass * self.pdf(mass, log_mode=True))
            else:
                return np.sum(d_log_mass * self.pdf(mass, log_mode=True))
        else:
            d_mass = (mass_max - mass_min) / float(steps)
            mass = np.linspace(mass_min, mass_max, steps)

            if weight:
                return np.sum(mass * d_mass * self.pdf(mass, log_mode=False))
            else:
                return np.sum(d_mass * self.pdf(mass, log_mode=False))

    def sample(self, n, mass_min=0.1, mass_max=10., steps=10000, seed=None):
        """
        Sample initial mass values between mass_min and mass_max,
        following the IMF distribution.

        ADW: Should this be `sample` or `simulate`?

        Parameters:
        -----------
        n : number of samples to draw
        mass_min : minimum mass to sample from
        mass_max : maximum mass to sample from
        steps    : number of steps for isochrone sampling
        seed     : random seed (passed to np.random.seed)

        Returns:
        --------
        mass     : array of randomly sampled mass values
        """
        if seed is not None: np.random.seed(seed)
        steps = int(steps)
        d_mass = (mass_max - mass_min) / float(steps)
        mass = np.linspace(mass_min, mass_max, steps)
        cdf = np.insert(np.cumsum(d_mass * self.pdf(mass[1:], log_mode=False)), 0, 0.)
        cdf = cdf / cdf[-1]
        f = scipy.interpolate.interp1d(cdf, mass)
        return f(np.random.uniform(size=n))

    @abstractmethod
    def pdf(cls, mass, **kwargs): pass
        

class Chabrier2003(IMF):
    """ Initial mass function from Chabrier (2003):
    "Galactic Stellar and Substellar Initial Mass Function"
    Chabrier PASP 115:763-796 (2003)
    https://arxiv.org/abs/astro-ph/0304382    
    """

    @classmethod
    def pdf(cls, mass, log_mode=True):
        """ PDF for the Chabrier IMF.
         
        The functional form and coefficients are described in Eq 17
        and Tab 1 of Chabrier (2003):

          m <= 1 Msun: E(log m) = A1*exp(-(log m - log m_c)^2 / 2 sigma^2)
          m  > 1 Msun: E(log m) = A2 * m^-x
         
          A1 = 1.58 : normalization [ log(Msun)^-1 pc^-3]
          m_c = 0.079 [Msun]
          sigma = 0.69
          A2 = 4.43e-2
          x = 1.3
          
        We redefine a = A1, A2 = a * b;

        The normalization is set so that the IMF integrates to 1 over
        the mass range from 0.1 Msun to 100 Msun

        Parameters:
        -----------
        mass: stellar mass (solar masses)
        log_mode[True]: return number per logarithmic mass range, i.e., dN/dlog(M)
         
        Returns:
        --------
        number per (linear or log) mass bin, i.e., dN/dM or dN/dlog(M) where mass unit is solar masses
        """
        log_mass = np.log10(mass)

        # Constants from Chabrier 2003
        m_c = 0.079
        sigma = 0.69
        x = 1.3 

        # This value is set to normalize integral from 0.1 to 100 Msun
        a=1.31357499301
        # This value is required so that the two components match at 1 Msun
        b = 0.279087531047 

        dn_dlogm = ((log_mass <= 0) * a * np.exp(-(log_mass - np.log10(m_c))**2 / (2 * (sigma**2)))) 
        dn_dlogm += ((log_mass  > 0) * a * b * mass**(-x))
        
        if log_mode:
            # Number per logarithmic mass range, i.e., dN/dlog(M)
            return dn_dlogm
        else:
            # Number per linear mass range, i.e., dN/dM
            return dn_dlogm / (mass * np.log(10))

class Kroupa2001(IMF):
    """ IMF from Kroupa (2001):

    "On the variation of the initial mass function"
    MNRAS 322:231 (2001)
    https://arxiv.org/abs/astro-ph/0009005
    """

    @classmethod
    def pdf(cls, mass, log_mode=True):
        """ PDF for the Kroupa IMF.

        Normalization is set over the mass range from 0.1 Msun to 100 Msun
        """
        log_mass = np.log10(mass)
        # From Eq 2
        mb = mbreak  = [0.08, 0.5] # Msun
        a = alpha = [0.3, 1.3, 2.3] # alpha

        # Normalization set from 0.1 -- 100 Msun
        norm = 0.27947743949440446

        b = 1./norm
        c = b * mbreak[0]**(alpha[1]-alpha[0])
        d = c * mbreak[1]**(alpha[2]-alpha[1])

        dn_dm  = b * (mass < 0.08) * mass**(-alpha[0])
        dn_dm += c * (0.08 <= mass) * (mass < 0.5) * mass**(-alpha[1])
        dn_dm += d * (0.5  <= mass) * mass**(-alpha[2])

        if log_mode:
            # Number per logarithmic mass range, i.e., dN/dlog(M)
            return dn_dm * (mass * np.log(10))
        else:
            # Number per linear mass range, i.e., dN/dM
            return dn_dm
    
class Salpeter1955(IMF): 
    """ IMF from Salpeter (1955):
    "The Luminosity Function and Stellar Evolution"
    ApJ 121, 161S (1955)
    http://adsabs.harvard.edu/abs/1955ApJ...121..161S
    """

    @classmethod
    def pdf(cls, mass, log_mode=True):
        """ PDF for the Salpeter IMF.

        Value of 'a' is set to normalize the IMF to 1 between 0.1 and 100 Msun
        """
        alpha = 2.35

        a = 0.060285569480482866
        dn_dm  = a * mass**(-alpha)

        if log_mode:
            # Number per logarithmic mass range, i.e., dN/dlog(M)
            return dn_dm * (mass * np.log(10))
        else:
            # Number per linear mass range, i.e., dN/dM
            return dn_dm
        
def factory(name, **kwargs):
    from ugali.utils.factory import factory
    return factory(name, module=__name__, **kwargs)

imfFactory = factory

############################################################

def chabrierIMF(mass, log_mode=True):
    """ Backward compatible wrapper around Chabrier2003.pdf """
    return Chabrier2003.pdf(mass,log_mode=log_mode)

############################################################
